unit uBafVarModule;

// this code is under the BAF fair use license (BFUL) - https://bafbal.de/index.php?title=Bful
// module for vars, text, csv...
// https://bafbal.de/index.php?title=Modul_VAR_neu

interface

uses System.Math, System.SysUtils, System.Classes, uBafTypes, uBafInterpreter,
  System.Contnrs, Winapi.Windows, System.StrUtils, FMX.Dialogs,
  System.Hash, uStringIniFile, System.RegularExpressions, System.NetEncoding;


type
  TBafNumInfo = (niIsInt, niIsCurr, niIntLen, niIntMin, niIntMax, niDateMin,
      niDateMax, niNumBetween, niDateBetween);

  TBafColorInfo = (ciDeadlineYR);

  TBafVarModule = class(TBafInterpreterCustomModule)
  protected   // Text
    FTextList: TObjectList;
    procedure SetText(AIndex: integer);
    procedure SetTextN(AIndex: integer);
    procedure SaveText(AIndex: integer);
    procedure OpenText(AIndex: integer);
    procedure TextProperties(AIndex: integer);
    procedure TextSet(AIndex: integer);
    procedure TextCompose(AIndex: integer);
    procedure TextDissect(AIndex: integer);
    procedure TextSort(AIndex: integer);
    procedure TextLoop(AIndex: integer);
    procedure ClearText(AIndex: integer);
    function GetText(AParam: string): string;
    function GetTextLine(AParams: TStrings): string;
  protected   // CSV
    FCsvList, FCsvHeaderList, FCsvZeileList: TObjectList;
    function GetCsvList(AIndex: integer): TStringList;
    function GetCsvHeaderList(AIndex: integer): TStringList;
    function GetCsvZeileList(AIndex: integer): TStringList;
    procedure CsvOpen(AIndex: integer);
    procedure CsvLine(AIndex: integer);
    function GetCsvValue(AIndex, AName: string): string;
  protected  // Ini
    FIniList: TObjectList;
    procedure SetIniValue(AIndex: integer);
    procedure IniLoadSave(AIndex: integer; ASave: boolean);
    function GetIniText(AParam: string): string;
    function GetIniValue(AParams: TStrings): string;
  protected   // Sperated Line
    FSepLineList: TObjectList;
    FSepLineValue: TStringList;
    function GetSepLineList(AIndex: integer): TStringList;
    procedure SepLine(AIndex: integer);
    function GetSepLine(AParam: string): string;
  protected   // Vars und Code
    FCode: TStringList;
    FNamedVars: TStringList;
    FCmdList: TObjectList;
    procedure SetVar(AOnlyIfEmpty: boolean);
    procedure AddVar;
    function GetVar(AParams: TStrings): string;
    function GetVariable(AName: string): string; override;
    procedure SetVariable(AName: string; const Value: string); override;
    function GetRightOfDef(AName: string): TBafRight; override;
    procedure LogVars;
    function GetCode(AClear: boolean): string; override;
    procedure SetCmd(AIndex: integer);
    procedure ClearCmd(AIndex: integer);
    function GetCmdList(AIndex: integer): TStringList;
    function GetProc(AIndex: integer): string; override;
    procedure ClearCmds;
  protected // Rights
    FRightDefs: TStringList;
    procedure RightsDef;
    function GetUserId(AParams: TStrings): string;
  protected    // Funktionen
    function Int2Time(AParam: string): string;
    function RemoveCrLf(AParams: TStrings): string;
    procedure Exec;
    procedure Procdir;
    procedure Forcedir;
    procedure BatchExec;
    procedure DoOpenFile;
    procedure Save2Clipboard;
    procedure FileOperation;
    procedure BafFileSearch;
    procedure PasteCsv;
    procedure Loop;
    procedure CheckString;
    function GetNumInfos(AInfo: TBafNumInfo; AParams: TStrings): string;
    function GetIntMinMax(AInfo: TBafNumInfo; ANum, ANum2: string): string;
    function GetDateMinMaxRel(AInfo: TBafNumInfo; ANum, ANum2: string): string;
    function NewGuidFormat(AText: string): string;
    function StringCopy(AParams: TStrings): string;
    function GetLookup(AParams: TStrings): string;
    function GetInversLookup(AParams: TStrings): string;
    function GetChar(AParams: TStrings): string;
    function BafBase64(AParams: TStrings): string;
    function IncDec(ADec: boolean; AParams: TStrings): string;
    function ICalc(AParams: TStrings): string;
    function BafPos(AParams: TStrings): string;
    function BafRandom(AParams: TStrings): string;
    function ReplaceString(AParams: TStrings): string;
    function StringExtract(AParams: TStrings): string;
    function BafTrim(AParams: TStrings): string;
    function BafXmlReplace(AParams: TStrings): string;
    function BafXmlReReplace(AParams: TStrings): string;
    function TextCompare(AParams: TStrings): string;
    function GetGuid(AParams: TStrings): string;
    function FunctionExecute(AParams: TStrings): string;
    function GetHash(AParams: TStrings): string;
    function NullValue(AParams: TStrings): string;
    function Pad(AParams: TStrings): string;
    function FormatDataTime(AParams: TStrings; ADefault: string): string;
    function LowNoSpace(AText: string): string;
    function GetClrDeadline(AInfo: TBafColorInfo; AParams: TStrings): string;
    function Convert(AParams: TStrings): string;
    function IncDate(AParams: TStrings): string;
    function GetDateFormated(AParams: TStrings): string;
    function FileDialog(AParams: TStrings): string;
    function IsAdmin(AParams: TStrings; ARight: boolean): string;
    function BCalc(AParams: TStrings): string;
    function CCalc(AParams: TStrings): string;
    function GetCurr(AParams: TStrings): string;
    function OnlyNum(AParams: TStrings): string;
    function InOperation(AParams: TStrings; AIgnoreCase: boolean): string;
    function GetBetween(AInfo: TBafNumInfo; AParams: TStrings): string;
    function TextHasLine(AParams: TStrings): string;
    function BafREgex(AParams: TStrings): string;
  protected
    function GetJSON(AParams: TStrings): string;
    function BafJsonObjectParse(AJson, APath: string;
        AIgnoreNameCase: boolean = false; APathSep: Char = '.'): string;
  public
    constructor Create; override;
    destructor Destroy; override;
    function InterpretLine(AExecInter: TBafCustomInterpreter): boolean; override;
    function ReplaceFunction(ACommand: string; AParams: TStrings; var AResult: string): boolean; override;
    function GetTextStringList(AIndex: integer): TStringList; override;
    function GetIni(AIndex: integer): TStringIniFile; override;
  end;

implementation

{ TBafVarModule }

uses uBafDataCache, uOsStuff, dmMain, uBafComboHelper;

procedure TBafVarModule.AddVar;
var
  LName, LValue, LRightName, LTyp: string;
  LRight: TBafRight;
begin
  if FindParamBooleanReplaced('cnd', true) then begin
    LRightName := FindParamStringLower('r', 'frm');
    LName := FindParamStringReplaced('n', '');
    if LRightName = 'frm' then
      LRight := brWrite
    else
      LRight := FInter.GetRightOfDef(LRightName);
    if LRight = brWrite then begin
      LValue := FindParamStringReplaced('z', '');
      LTyp := FindParamStringReplacedLower('y', '');
      if LTyp = 'int' then
        FNamedVars.Values[LName] := IntToStr(StrToIntDef(FNamedVars.Values[LName], 0)
            + StrToIntDef(LValue, 0))
      else if LTyp = 'curr' then
        FNamedVars.Values[LName] := CurrToStr(StrToCurrDef(FNamedVars.Values[LName], 0)
            + StrToCurrDef(LValue, 0))
      else if LTyp = 'date' then
        FNamedVars.Values[LName] := FormatDateTime('dd.mm.yyyy',
            StrToDateTimeDef(FNamedVars.Values[LName], 0) + StrToFloatDef(LValue, 0))
      else if LTyp = 'datetime' then begin
        FNamedVars.Values[LName] := FormatDateTime('dd.mm.yyyy hh:mm:ss',
            StrToDateTimeDef(FNamedVars.Values[LName], 0) + StrToFloatDef(LValue, 0))
        end
      else
        FNamedVars.Values[LName] := FNamedVars.Values[LName] + LValue;
    end;
  end;
end;

function TBafVarModule.BafBase64(AParams: TStrings): string;
var
  s0, s1: string;
begin
  if AParams.Count > 1 then begin
    s0 := AnsiLowerCase(AParams[0]) + ' ';
    s1 := AParams[1];
    if s0[1] = 'e' then
      result := TNetEncoding.Base64String.Encode(s1)
    else if s0[1] = 'd' then
      result := TNetEncoding.Base64String.Decode(s1);
  end
  else
    FInter.DoLog('E', '$BASE64 - Number of Params less 2');
end;

procedure TBafVarModule.BafFileSearch;
var
  LText: TStringList;
  LNum, LResult, LAttr: integer;
  LSearchRec: TSearchRec;
  LFileName, LAttributes: string;
  LAll: boolean;
begin
  if FindParamBooleanReplaced('cnd', true) then begin
    LNum := FindParamIntegerReplaced('n', 1);
    LText := GetTextStringList(LNum);
    LFileName := FindParamStringReplaced('fn', '');
    LAll := FindParamBooleanReplaced('all', false);
    LAttributes := FindParamStringReplacedLower('y', 'faAnyFile');
    if LAttributes = 'AnyFile' then
      LAttr := faAnyFile
    else if LAttributes = 'ReadOnly' then
      LAttr := faReadOnly
    else if LAttributes = 'Hidden' then
      LAttr := faHidden
    else if LAttributes = 'SysFile' then
      LAttr := faSysFile
    else if LAttributes = 'VolumeID' then
      LAttr := faVolumeID
    else if LAttributes = 'Directory' then
      LAttr := faDirectory
    else if LAttributes = 'Archive' then
      LAttr := faArchive
    else LAttr := StrToIntDef(LAttributes, faAnyFile);
    if FindParamBooleanReplaced('clr', true) then
      LText.Clear;
    LResult := FindFirst(LFileName, LAttr, LSearchRec);
    try
       while LResult = 0 do begin
         if LAll then
           LText.Add(Format('%s;%s;%d;%d', [LSearchRec.Name,
               FormatDateTime(BAFDATETIMEFMT, (LSearchRec.TimeStamp)),
               LSearchRec.Size, LSearchRec.Attr]))
         else
           LText.Add(LSearchRec.Name);
         LResult := FindNext(LSearchRec);
       end;
    finally
      System.SysUtils.FindClose(LSearchRec);
    end;
  end;
end;

function TBafVarModule.BafJsonObjectParse(AJson, APath: string;
  AIgnoreNameCase: boolean; APathSep: Char): string;
var
  i, LLevel, LLvlObject, LLvlArray, LLvlObject2: integer;
  LName, LValue, LParsePath: string;
  LAbort, LInQuote, LInValue, LInObject: boolean;

  procedure lokInit;
  begin
    APath := APathSep + APath;
    LLevel := 0;
    LLvlObject := -1;
    LLvlArray := 0;
    LInQuote := false;
    LInValue := false;
    LAbort := false;
    LInObject := false;
    LLvlObject2 := 0;
  end; // procedure lokInit

  procedure lokReducePath;
  var
    i: integer;
  begin
    for i := Length(LParsePath) downto 1 do begin
      if LParsePath[i] = APathSep then begin
        Delete(LParsePath, i, MaxInt);
        exit;
      end;
    end;
  end; // procedure lokReducePath

  procedure lokStopValue;
  begin
    LInValue := false;
    if AIgnoreNameCase then begin
      if AnsiCompareText(LParsePath + APathSep + LName, APath) = 0 then begin
        result := LValue;
        LAbort := true;
      end;
    end  // if AIgnoreNameCase
    else begin
      if LParsePath + APathSep + LName = APath then begin
        result := LValue;
        LAbort := true;
      end;
    end; // else AIgnoreNameCase
    LName := '';
  end; // procedure lokStartValue

  procedure lokCheckCharIntern;
  begin
    case AJson[i] of
      '"': LInQuote := not LInQuote;
      '{': if not LInQuote then begin
             inc(LLvlObject);
             if LLvlObject > 0 then
               LParsePath := LParsePath + APathSep + LName;
             if AIgnoreNameCase then begin
               if AnsiCompareText(LParsePath, APath) = 0 then
                 LInObject := true;
             end
             else begin
               if LParsePath = APath then
                 LInObject := true;
             end;
             LInValue := false;
             LName := '';
           end;
      '}': if not LInQuote then begin
             dec(LLvlObject);
             lokStopValue;
             lokReducePath;
           end;
      '[': if not LInQuote then begin
             inc(LLvlArray);
             LValue := LValue + '[';
           end;
      ':': if not LInQuote then begin
             LInValue := true;
             LValue := '';
           end;
      ',': if not LInQuote then
             lokStopValue;
      ' ': if LInQuote then begin
             if LInValue then
               LValue := LValue + AJson[i]
             else
               LName := LName + AJson[i];
           end;
      else begin
        if LInValue then
          LValue := LValue + AJson[i]
        else if LInQuote then
          LName := LName + AJson[i];
      end;
    end; // case
  end; // procedure lokCheckCharIntern

  procedure lokCheckChar;
  begin
    if not LInObject then begin
      if LLvlArray = 0 then
        lokCheckCharIntern
      else begin
        if AJson[i] = ']' then
          dec(LLvlArray);
        LValue := LValue + AJson[i];
        if LLvlArray = 0 then
          lokStopValue;
      end;
    end
    else begin  // if not LInObject
      case AJson[i] of
        '"': LInQuote := not LInQuote;
        '{': if not LInQuote then
               inc(LLvlObject2);
        '}': if not LInQuote then
               dec(LLvlObject2);
      end; // case AJson[i] of
      LValue := LValue + AJson[i];
      if LLvlObject2 < 0 then begin
        result := '{' + LValue;
        LAbort := true;
      end;
    end;  // else if not LInObject
  end; // procedure lokCheckChar

begin
  lokInit;
  for i := 1 to Length(AJson) do begin
    lokCheckChar;
    if LAbort then
      Break;
  end;
// function TBafVarModule.BafJsonObjectParse
end;

function TBafVarModule.BafPos(AParams: TStrings): string;
// 0 - String to search
// 1 - String to seachr in
// 2 - Options
var
  s0, s1: string;
begin
  if AParams.Count > 1 then begin
    s0 := AParams[0];
    s1 := AParams[1];
    if AParams.Count > 2 then begin
      if AnsiCompareText(AParams[2], 'ic') = 0 then begin
        s0 := AnsiUpperCase(s0);
        s1 := AnsiUpperCase(s1);
      end;
    end;
    result := IntToStr(Pos(s0, s1));
  end
  else
    FInter.DoLog('E', '$POS - Number of Params less 2');
end;

function TBafVarModule.BafRandom(AParams: TStrings): string;
var
  LTyp, LValueMin, LValueMax: string;
  LIntMin, LIntMax, LTextNum: integer;
  LDateMin, LDateMax: TDate;
  LCurrMin, LCurrMax: currency;
  LText: TStringList;
  LHelper: TBafComboHelper;
begin
  if AParams.Count > 0 then begin
    LTyp := AnsiLowerCase(AParams[0]);
    if AParams.Count > 1 then
      LValueMin := AParams[1];
    if AParams.Count > 2 then
      LValueMax := AParams[2];

    if LTyp = 'int' then begin
      LIntMin := StrToIntDef(LValueMin, 0);
      LIntMax := StrToIntDef(LValueMax, 100);
      result := IntToStr(random(1 + LIntMax - LIntMin) + LIntMin);
    end
    else if LTyp = 'date' then begin
      LDateMin := StrToDateDef(LValueMin, now - 7);
      LDateMax := StrToDateDef(LValueMax, now + 7);
      result := FormatDateTime(BAFDATEFMT, (random(1
          + round(LDateMax) - round(LDateMin)) + round(LDateMin)));
    end
    else if LTyp = 'curr' then begin
      LCurrMin := StrToCurrDef(LValueMin, 0);
      LCurrMax := StrToCurrDef(LValueMax, 100);
      result := FormatFloat('0.00', (LCurrMax - LCurrMin) * random + LCurrMin);
    end
    else if LTyp = 'curr4' then begin
      LCurrMin := StrToCurrDef(LValueMin, 0);
      LCurrMax := StrToCurrDef(LValueMax, 100);
      result := CurrToStr((LCurrMax - LCurrMin) * random + LCurrMin);
    end
    else if copy(LTyp, 1, 4) = 'text' then begin
      LTextNum := StrToIntDef(copy(LTyp, 5, MaxInt), 1);
      LText := GetTextStringList(LTextNum);
      result := LText[random(LText.Count)];
    end
    else if LTyp = 'ld' then begin
      if LValueMin <> '' then begin
        LHelper := gvBafDataCache.Lookup[LValueMin];
        result := LHelper.GetGuidFromIndex(random(LHelper.Count));
      end;
    end
    else if LTyp = 'ldv' then begin
      if LValueMin <> '' then begin
        LHelper := gvBafDataCache.Lookup[LValueMin];
        result := LHelper.Strings[random(LHelper.Count)];
      end;
    end
    else if LTyp = 'ls' then begin
      if LValueMin <> '' then begin
        LHelper := gvBafDataCache.Special[LValueMin];
        result := LHelper.GetGuidFromIndex(random(LHelper.Count));
      end;
    end
    else if LTyp = 'lsv' then begin
      if LValueMin <> '' then begin
        LHelper := gvBafDataCache.Special[LValueMin];
        result := LHelper.Strings[random(LHelper.Count)];
      end;
    end;

  end;
// function TBafVarModule.BafRandom
end;

function TBafVarModule.BafREgex(AParams: TStrings): string;
begin
  if AParams.Count > 1 then begin
    result := IfThen(TRegEx.IsMatch(Trim(AParams[0]), Trim(AParams[1])), 'Y', 'N');
  end
  else
    FInter.DoLog('E', '$REGEX - Number of Params less 2');
end;

function TBafVarModule.BafTrim(AParams: TStrings): string;
var
  LTrim: string;
begin
  if AParams.Count > 0 then begin
    if AParams.Count > 1 then
      LTrim := AnsiUpperCase(AParams[1]);
    case (LTrim + ' ')[1] of
      'L': result := TrimLeft(AParams[0]);
      'R': result := TrimRight(AParams[0]);
    else
      result := Trim(AParams[0]);
    end;
  end
  else
    FInter.DoLog('E', '$TRIM - Number of Params less 1');
end;

function TBafVarModule.BafXmlReplace(AParams: TStrings): string;
var
  i: integer;
  s: string;
begin
  s := AParams[0];
  result := '';
  for i := 1 to Length(s) do begin
    case s[i] of
      '&': result := result + '&amp;';
      '<': result := result + '&lt;';
      '>': result := result + '&gt;';
      '"': result := result + '&quot;';
      Chr(39): result := result + '&apos;';
      else
        result := result + s[i];
    end;
  end;
end;

function TBafVarModule.BafXmlReReplace(AParams: TStrings): string;
var
  i: integer;
  s: string;
begin
  s := AParams[0];
  result := '';
  i := 1;
  while i <= Length(s) do begin
    if s[i] = '&' then begin
      if copy(s, i, 5) = '&amp;' then begin
        result := result + '&';
        inc(i, 4);
      end
      else if copy(s, i, 4) = '&lt;' then begin
        result := result + '<';
        inc(i, 3);
      end
      else if copy(s, i, 4) = '&gt;' then begin
        result := result + '>';
        inc(i, 3);
      end
      else if copy(s, i, 6) = '&quot;' then begin
        result := result + '"';
        inc(i, 5);
      end
      else if copy(s, i, 6) = '&apos;' then begin
        result := result + Chr(39);
        inc(i, 5);
      end
      else
        result := result + s[i];
    end
    else
      result := result + s[i];
    inc(i);
  end;
end;

procedure TBafVarModule.BatchExec;
// Speichert den Text n in batch.bat und fhrt diese aus
var
  LFileName: string;
  LIndex: integer;
  LText: TStringList;
begin
  LFileName := FInter.NeedInfo('usrroot', '') + 'batch.bat';
  LIndex := FindParamInteger('n', 0);
  LText := GetTextStringList(LIndex);
  LText.SaveToFile(LFileName);
  BafOpenFile(LFileName);
  if FindParamBooleanReplaced('clr', true) then
    LText.Clear;
end;

function TBafVarModule.BCalc(AParams: TStrings): string;
// 0 - operator
// 1 - value 1
// 2 - value 2
// optional 3 and more: further value
var
  LOp: string;
  LResult: boolean;
  i: integer;
begin
  if AParams.Count > 2 then begin
    LOp := AnsiLowerCase(AParams[0]);
    LResult := false;
    if LOp = 'and' then begin
      LResult := true;
      for i := 1 to AParams.Count - 1 do begin
        if not BafIsYesChar(AParams[i]) then begin
          LResult := false;
          Break;
        end;
      end;
    end
    else if LOp = 'or' then begin
      LResult := false;
      for i := 1 to AParams.Count - 1 do begin
        if BafIsYesChar(AParams[i]) then begin
          LResult := true;
          Break;
        end;
      end;
    end
    else if LOp = 'xor' then
      LResult := BafIsYesChar(AParams[1]) xor BafIsYesChar(AParams[2]);
    result := IfThen(LResult, BAFYESCHAR, BAFNOCHAR);
  end
  else
    FInter.DoLog('E', '$BCALC - number of params too low');
end;

function TBafVarModule.CCalc(AParams: TStrings): string;
// 0 - operator
// 1 - number 1
// 2 - numer 2
// optional 3 and more: further numbers
var
  LOp: string;
  LResult: currency;
  i: integer;
begin
  if AParams.Count > 2 then begin
    LOp := AParams[0] + '2';
    LResult := 0;
    for i := 1 to AParams.Count - 1 do
      AParams[i] := StringReplace(AParams[i], '.', ',', []);
    if LOp[1] = '+' then begin
      for i := 1 to AParams.Count - 1 do
        LResult := LResult + StrToCurr(AParams[i]);
    end
    else if LOp[1] = '*' then begin
      LResult := 1;
      for i := 1 to AParams.Count - 1 do
        LResult := LResult * StrToCurr(AParams[i]);
    end
    else if LOp[1] = '-' then begin
      LResult := StrToCurr(AParams[1]);
      for i := 2 to AParams.Count - 1 do
        LResult := LResult - StrToCurr(AParams[i]);
    end
    else if LOp[1] = '/' then begin
      LResult := StrToCurr(AParams[1]);
      for i := 2 to AParams.Count - 1 do
        LResult := LResult / StrToCurr(AParams[i]);
    end
    else if LOp[1] = '%' then begin
      LResult := 100 * StrToCurr(AParams[1]);
      for i := 2 to AParams.Count - 1 do
        LResult := LResult / StrToCurr(AParams[i]);
    end;
    if LOp[2] = '4' then
      result := FormatCurr('0.0000', LResult)
    else
      result := FormatCurr('0.00', LResult);
  end
  else
    FInter.DoLog('E', '$ICALC - number of params too low');
end;

procedure TBafVarModule.CheckString;
var
  LValue, LTyp, LSuc, LMsg, LMessage: string;
  LSuccess: boolean;
  LNum: integer;

  procedure lokTtDeRp;
  var
    i: integer;
    s, t: string;
  begin
    if Length(LValue) <> 9 then begin
      LMessage := Format('Lnge %d Zeichen - ', [Length(LValue)]);
      LSuccess := false;
    end;
    if (Length(LValue) > 0) and (LValue[1] <> 'C') then begin
      LMessage := LMessage + 'Erstes Zeichen nicht "C" - ';
      LSuccess := false;
    end;
    s := '';
    t := '';
    for i := 1 to Length(LValue) do begin
      if (LValue[i] = AnsiLowerCase(LValue[i])) and not CharInSet(LValue[i], ['0'..'9']) then begin
        s := s + ', ' + LValue[i];
        LSuccess := false;
      end
      else if not CharInSet(LValue[i], ['0'..'9', 'C', 'F', 'G', 'H', 'J', 'K', 'L', 'M',
          'N', 'P', 'R', 'T', 'V', 'W', 'X', 'Y', 'Z']) then begin
        t := t + ', ' + LValue[i];
        LSuccess := false;
      end;
    end;
    if Length(s) > 0 then
      LMessage := LMessage + 'Zeichen in Kleinschreibung: ' + copy(s, 3, MaxInt) + ' - ';
    if Length(t) > 0 then
      LMessage := LMessage + 'Unzulssige Zeichen: ' + copy(t, 3, MaxInt) + ' - ';
  end; // procedure lokTtDeRp

begin
  if FindParamBooleanReplaced('cnd', true) then begin
    LValue := FindParamStringReplaced('z', '');
    LTyp := FindParamStringReplacedLower('y', '');
    LSuc := FindParamStringReplaced('suc', '');
    LMsg := FindParamStringReplaced('msg', '');
    LSuccess := true;
    if LTyp = 'tt_de_rp' then
      lokTtDeRp
    ;
    LNum := StrToIntDef(LSuc, -1);
    if LNum > 0 then
      FExecInter.Values[LNum] := IfThen(LSuccess, BAFYESCHAR, BAFNOCHAR)
    else if LSuc <> '' then
      FInter.Variable[LSuc] := IfThen(LSuccess, BAFYESCHAR, BAFNOCHAR);
    LNum := StrToIntDef(LMsg, -1);
    if LNum > 0 then
      FExecInter.Values[LNum] := LMessage
    else if LSuc <> '' then
      FInter.Variable[LMsg] := LMessage;
  end;
end;

procedure TBafVarModule.ClearCmd(AIndex: integer);
begin
  GetCmdList(AIndex).Clear;
  if Trim(FExecInter.LineP) <> '' then
    GetCmdList(AIndex).Add(FExecInter.LineP);
end;

procedure TBafVarModule.ClearCmds;
begin
  FCmdList.Clear;
end;

procedure TBafVarModule.ClearText(AIndex: integer);
var
  s: string;
begin
  GetTextStringList(AIndex).Clear;
  s := FExecInter.ReplaceFunctions(FExecInter.LineP);
  s := FExecInter.ReplaceFunctions(s);
  if Trim(s) <> '' then
    GetTextStringList(AIndex).Add(s);
end;

function TBafVarModule.Convert(AParams: TStrings): string;
// 1 - Konvertierungsfunktion
// 2 - Konvertierungsdaten
var
  LFunction: string;
begin
  LFunction := AnsiLowerCase(AParams[0]);
  if LFunction = 'isodate' then
    result := BafConvert(cfIsoDate, AParams[1])
  else if LFunction = 'isodatetime' then
    result := BafConvert(cfIsoDateTime, AParams[1])
  else if LFunction = 'truefalse' then
    result := BafConvert(cfTrueFalse, AParams[1])
  else if LFunction = 'pointfloat' then
    result := BafConvert(cfPointFloat, AParams[1])
  else
    result := AParams[1];
end;

constructor TBafVarModule.Create;
begin
  inherited;
  FTextList := TObjectList.Create(true);
  FCsvList := TObjectList.Create(true);
  FIniList := TObjectList.Create(true);
  FCsvHeaderList := TObjectList.Create(true);
  FCsvZeileList := TObjectList.Create(true);
  FSepLineList := TObjectList.Create(true);
  FSepLineValue := TStringList.Create;
  FNamedVars := TStringList.Create;
  FCode := TStringList.Create;
  FCmdList := TObjectList.Create(true);
  FRightDefs := TStringList.Create;
end;

procedure TBafVarModule.CsvLine(AIndex: integer);
var
  LRow, LFirst, LRowCount, LMax: integer;
  LBafConName, LSep, LEachRow: string;
  LEachRowTrans, LNoException: boolean;
  LText, LHeaderText, LZeileText: TStringList;
begin
  LText := GetCsvList(AIndex);
  LHeaderText := GetCsvHeaderList(AIndex);
  LZeileText := GetCsvZeileList(AIndex);
  LSep := FindParamString('sep', ';') + ' ';
  LZeileText.Delimiter := LSep[1];
  LZeileText.StrictDelimiter := true;
  LHeaderText.Delimiter := LSep[1];
  LMax := FindParamInteger('m', MaxInt);
  LEachRow := FindParamStringReplaced('er', '');
  LEachRowTrans := FindParamBoolean('ert', false);
  LNoException := FindParamBooleanReplaced('nex', false);
  LBafConName := FindParamStringReplacedLower('db', DC_DEFAULT);
  if (LEachRow <> '') and (LText.Count > 0) then begin
    LRowCount := 0;
    LHeaderText.Clear;
    LFirst := 0;
    if FindParamBoolean('hhr', false) then begin
      LFirst := 1;
      LHeaderText.DelimitedText := AnsiLowerCase(LText[0]);
    end;

    for LRow := LFirst to LText.Count - 1 do begin
      LZeileText.DelimitedText := LText[LRow];
      FExecInter.EachRow(LBafConName, LEachRow, '#csvline', LEachRowTrans,
          LNoException);
      inc(LRowCount);
      if LRowCount >= LMax then begin
        FInter.DoLog('I', Format('#csvline, Max (%d) reached, execution aborted', [LMax]));
        Break;
      end;
    end;
  end;
// procedure TBafVarModule.CsvLine
end;

procedure TBafVarModule.CsvOpen(AIndex: integer);
var
  LFileName, s: string;
  i, LTextNum: integer;
  LInQuote: boolean;
  LText: TStringList;
begin
  LText := GetCsvList(AIndex);
  LFileName := FindParamStringReplaced('fn', '');
  LTextNum := FindParamIntegerReplaced('txt', 0);
  if FileExists(LFileName) then
    LText.LoadFromFile(LFileName)
  else if LTextNum > 0 then
    LText.Text := GetTextStringList(LTextNum).Text
  else
    FInter.DoLog('E', '#csvopen - Datei existiert nicht: ' + LFileName);

  if FindParamBoolean('ber', false) then begin    // replace line breaks in fields through blanks
    s := LText.Text;
    LInQuote := false;
    for i := 1 to Length(s) do begin
      if s[i] = '"' then
        LInQuote := not LInQuote;
      if LInQuote and CharInSet(s[i], [#13, #10]) then
        s[i] := ' ';
    end;
    LText.Text := s;
  end;

end;

destructor TBafVarModule.Destroy;
begin
  FreeandNil(FRightDefs);
  FreeandNil(FCmdList);
  FreeandNil(FCode);
  FreeandNil(FNamedVars);
  FreeandNil(FSepLineList);
  FreeandNil(FSepLineValue);
  FreeandNil(FCsvZeileList);
  FreeandNil(FCsvHeaderList);
  FreeandNil(FIniList);
  FreeandNil(FCsvList);
  FreeandNil(FTextList);
  inherited;
end;

procedure TBafVarModule.DoOpenFile;
var
  LFileName: string;
begin
  LFileName := FindParamStringReplaced('fn', '');
  if LFileName = '' then
    FInter.DoLog('E', '#openfile, fn: ' + LFileName)
  else
    BafOpenFile(LFileName);
end;

procedure TBafVarModule.Exec;
var
  LCmd: string;
begin
  if FindParamBooleanReplaced('cnd', true) then begin
    LCmd := FindParamStringReplaced('cmd', '');
    TBafInterpreterLevel.ExecInNewLevel(LCmd, FExecInter, FInter);
  end;
end;

function TBafVarModule.GetBetween(AInfo: TBafNumInfo; AParams: TStrings): string;
var
  LValue, LValue1, LValue2: double;
  i: integer;

  function lokGetValue(AIndex: integer): double;
  begin
    result := 0;
    case AInfo of
      niNumBetween: result := StrToFloatDef(AParams[AIndex], 0);
      niDateBetween: result := StrToDateTimeDef(AParams[AIndex], 0);
    end;
  end; // function lokGetValue

begin
  if AParams.Count > 2 then begin
    LValue := lokGetValue(0);
    LValue1 := lokGetValue(1);
    LValue2 := lokGetValue(2);
    result := IfThen((LValue >= LValue1) and (LValue <= LValue2),
        BAFYESCHAR, BAFNOCHAR);
    if result = BAFYESCHAR then
      exit;
    for i := 3 to AParams.Count -1 do begin
      LValue1 := lokGetValue(i);
      if LValue = LValue1 then begin
        result := BAFYESCHAR;
        exit;
      end;
    end;
  end
  else begin
    case AInfo of
      niNumBetween: FInter.DoLog('E', '$NUMBETWEEN - less then 3 parameters');
      niDateBetween: FInter.DoLog('E', '$DATEBETWEEN - less then 3 parameters');
    end;
  end;
end;

function TBafVarModule.GetChar(AParams: TStrings): string;
var
  s: string;
begin
  result := '';
  if AParams.Count > 0 then begin
    s := AnsiLowerCase(AParams[0]);
    if (s = 'crlf') or (s = 'lfcr') then
      result := #13#10
    else if s = 'cr' then
      result := #13
    else if s = 'cr' then
      result := #10
    else if s = 'lf' then
      result := #13
    else if s = 'tab' then
      result := #9
    else if s = 'quote' then
      result := #39
    else if s = 'dquote' then
      result := #34
    else if (s = 'space') or (s = 'leer') then
      result := ' '
    else if (s = 'comma') or (s = 'komma') then
      result := ','
    else if (s = 'questionmark') or (s = 'qmark') or (s = 'frage')
        or (s = 'fragezeichen') then
      result := '?'
    else if (s = 'bro') or (s = 'bracket_open')  then
      result := '('
    else if (s = 'brc') or (s = 'bracket_close')  then
      result := ')'
    else if s = 'dollar' then
      result := '$'

    else
      result := Chr(StrToIntDef(s, 0));
  end;
end;

function TBafVarModule.GetClrDeadline(AInfo: TBafColorInfo; AParams: TStrings): string;
// 0 - ceadline
// 1 - date to compare with deadline
// 2 - days before deadline "yellow"
var
  LDate, LDeadline: TDate;
  LGelb: integer;
  s, LFarbeGelb, LFarbeRot: string;
begin
  if AParams.Count < 3 then
    result := 'aqua'
  else begin
    LDeadline := StrToDateDef(AParams[0], 0);
    s := Trim(AnsiUpperCase(AParams[1]));
    if (Length(s) = 1) and CharInSet(s[1], ['H', 'T']) then
      LDate := trunc(now)
    else if (Length(s) = 1) and CharInSet(s[1], ['M']) then
      LDate := trunc(now) + 1
    else if (Length(s) = 1) and CharInSet(s[1], ['G', 'Y']) then
      LDate := trunc(now) - 1
    else
      LDate := StrToDateDef(s, 0);
    LGelb := StrToIntDef(AParams[2], 7);
    case AInfo of
      ciDeadlineYR: begin
        LFarbeGelb := 'yellow';
        LFarbeRot := 'red';
      end;
    end;
    result := '';
    if LDate >= LDeadline then
      result := LFarbeRot
    else if LDate >= (LDeadline - LGelb) then
      result := LFarbeGelb;
  end;
end;

function TBafVarModule.GetCode(AClear: boolean): string;
begin
  result := StringReplace(FCode.Text, #13#10, '   ', [rfReplaceAll]);
  if AClear then
    FCode.Clear;
end;

function TBafVarModule.GetCsvHeaderList(AIndex: integer): TStringList;
begin
  while AIndex > FCsvHeaderList.Count do
    FCsvHeaderList.Add(TStringList.Create);
  result := (FCsvHeaderList[AIndex - 1] as TStringList);
end;

function TBafVarModule.GetCsvList(AIndex: integer): TStringList;
begin
  while AIndex > FCsvList.Count do
    FCsvList.Add(TStringList.Create);
  result := (FCsvList[AIndex - 1] as TStringList);
end;

function TBafVarModule.GetCsvValue(AIndex, AName: string): string;
var
  LZeile, LHeader: TStringList;
  ix, LCol: integer;
begin
  ix := StrToIntDef(AIndex, -1);
  if ix >= 0 then begin
    result := '';
    LZeile := GetCsvZeileList(ix);
    LHeader := GetCsvHeaderList(ix);
    LCol := StrToIntDef(AName, -1);
    if LCol = -1 then begin
      LCol := LHeader.IndexOf(AnsiLowerCase(Trim(AName)));
      if (LCol < LZeile.Count) and (LCol >= 0) then
        result := LZeile[LCol]
      else
        FInter.DoLog('E', '$CSV(ix, n) - Column n not found: ' + AName);
    end
    else begin
      if LCol < LZeile.Count then
        result := LZeile[LCol]
      else
        FInter.DoLog('E', '$CSV(ix, n) - n must be less then the number of columns');
    end;
  end
  else
    FInter.DoLog('E', '$CSV(ix, n) - ix missing or lower 0');
end;

function TBafVarModule.GetCsvZeileList(AIndex: integer): TStringList;
begin
  while AIndex > FCsvZeileList.Count do
    FCsvZeileList.Add(TStringList.Create);
  result := (FCsvZeileList[AIndex - 1] as TStringList);
end;

function TBafVarModule.GetCurr(AParams: TStrings): string;
begin
  if AParams.Count > 0 then
    result := AParams[0];
  if AParams.Count > 1 then
    result := result + '.' + AParams[1];
end;

function TBafVarModule.GetDateFormated(AParams: TStrings): string;
var
  LDate: TDateTime;
begin
  if AParams.Count > 1 then begin
    LDate := StrToDateTimeDef(AParams[0], 0);
    result := FormatDateTime(AParams[1], LDate);
  end
  else
    FInter.DoLog('E', '$DFMT(date, fmt) - number of params low');
end;

function TBafVarModule.GetDateMinMaxRel(AInfo: TBafNumInfo; ANum, ANum2: string): string;
// compares the date in ANum with the current date, added with Num2
var
  LDate, LDate2: TDateTime;
begin
  result := '';
  LDate := StrToDateTimeDef(ANum, 0);
  LDate2 := now + StrToIntDef(ANum2, 0);
  case AInfo of
    niDateMin: result := IfThen(LDate < LDate2, BAFNOCHAR, BAFYESCHAR);
    niDateMax: result := IfThen(LDate > LDate2, BAFNOCHAR, BAFYESCHAR);
  end;
end;

function TBafVarModule.GetGuid(AParams: TStrings): string;
begin
  result := BafGetGuid;
  if AParams.Count > 0 then
    FInter.Variable[AParams[0]] := result;
end;

function TBafVarModule.GetHash(AParams: TStrings): string;
var
  LValue, LHash: string;
begin
  if AParams.Count > 1 then begin
    LValue := AParams[0];
    LHash := AnsiUpperCase(APArams[1]);
    if (LHash = 'SHA512') then
      result := THashSHA2.GetHashString(LValue, THashSHA2.TSHA2Version.SHA512)
    else if (LHash = 'SHA512_256') then
      result := THashSHA2.GetHashString(LValue, THashSHA2.TSHA2Version.SHA512_256)
    else if (LHash = 'SHA512_224') then
      result := THashSHA2.GetHashString(LValue, THashSHA2.TSHA2Version.SHA512_224)
    else if (LHash = 'SHA384') then
      result := THashSHA2.GetHashString(LValue, THashSHA2.TSHA2Version.SHA384)
    else if (LHash = 'SHA256') then
      result := THashSHA2.GetHashString(LValue, THashSHA2.TSHA2Version.SHA256)
    else if (LHash = 'SHA224') then
      result := THashSHA2.GetHashString(LValue, THashSHA2.TSHA2Version.SHA224)
    else if (LHash = 'SHA1') then
      result := THashSHA1.GetHashString(LValue)
    else if (LHash = 'MD5') then
      result := THashMD5.GetHashString(LValue);
  end;
end;

function TBafVarModule.GetIni(AIndex: integer): TStringIniFile;
begin
  while AIndex > FIniList.Count do
    FIniList.Add(TStringIniFile.Create(''));
  result := (FIniList[AIndex - 1] as TStringIniFile);
end;

function TBafVarModule.GetIniText(AParam: string): string;
var
  ix: integer;
begin
  ix := StrToIntDef(AParam, -1);
  if ix >= 0 then
    result := GetIni(ix).AsString
  else if AnsiCompareText(AParam, 'root') = 0 then
    result := dataMain.Ini.AsString
  else if AnsiCompareText(AParam, 'usr') = 0 then
    result := dataMain.UserIni.AsString
  else
    FInter.DoLog('E', '$INITEXT(ix) - ix incorrect');
end;

function TBafVarModule.GetIniValue(AParams: TStrings): string;
// 0 - Name or Index of the ini
// 1 - item
// 2 - optional: section; default values
// 3 - optional: default value
var
  ix: integer;
  LIniName: string;
  LIni: TStringIniFile;

  function lokVal(AIndex: integer; LDefault: string): string;
  begin
    if AParams.Count > AIndex then
      result := AParams[AIndex]
    else
      result := LDefault;
    if result = 'db!' then
      result := dataMain.DBName;
  end;

begin
  if AParams.Count > 1 then begin
    LIni := nil;
    LIniName := AnsiLowerCase(AParams[0]);
    if LIniName = 'usr' then
      LIni := dataMain.UserIni
    else if LIniName = 'root' then
      LIni := dataMain.Ini
    else begin
      ix := StrToIntDef(LIniName, -1);
      if ix >= 0 then
        LIni := GetIni(ix);
    end;
    if Assigned(LIni) then
      result := LIni.ReadString(lokVal(2, 'values'), AParams[1], lokVal(3, ''))
    else
      FInter.DoLog('E', '$INI(ix, item, [sec], [default]) - ini not found');
  end
  else
    FInter.DoLog('E', '$INI(ix, item, [sec], [default]) - count of params less 2');
end;

function TBafVarModule.GetIntMinMax(AInfo: TBafNumInfo; ANum, ANum2: string): string;
var
  LInt, LInt2: integer;
begin
  result := '';
  LInt := StrToIntDef(ANum, MaxInt);
  LInt2 := StrToIntDef(ANum2, MaxInt);
  case AInfo of
    niIntMin: result := IfThen(LInt < LInt2, BAFNOCHAR, BAFYESCHAR);
    niIntMax: result := IfThen(LInt > LInt2, BAFNOCHAR, BAFYESCHAR);
  end;
end;

function TBafVarModule.GetInversLookup(AParams: TStrings): string;
// Params
// 0 - name of the looup
// 1 - value
// 2 - default result
var
  LHelper: TBafComboHelper;
begin
  if AParams.Count > 1 then begin
    LHelper := gvBafDataCache.Lookup[AParams[0]];
    result := LHelper.GetGuidFromCaption(AParams[1]);
    if (result = '') and (AParams.Count > 2) then
      result := AParams[2];
  end
  else
    FInter.DoLog('E', '$INVLOOKUP(name, value, default) - number of params less 2');
end;

function TBafVarModule.GetJSON(AParams: TStrings): string;
var
  LVar, LPath, LText: string;
  LIgnoreCase: boolean;
  LSep: Char;
begin
  if AParams.Count > 1 then begin
    LPath := AParams[1];
    LVar := AParams[0];
    LText := FInter.Variable[LVar];
    LIgnoreCase := false;
    LSep := '.';
    if AParams.Count > 2 then
      LIgnoreCase := BafIsYesChar(AParams[2]);
    if AParams.Count > 3 then
      LSep := (AParams[2] + '.')[1];
    result := BafJsonObjectParse(LText, LPath, LIgnoreCase, LSep);
  end
  else
    FInter.DoLog('E', '$JSON, number of params less 2');
end;

function TBafVarModule.GetLookup(AParams: TStrings): string;
// Params
// 0 - name of the looup
// 1 - value
// 2 - default result
var
  LHelper: TBafComboHelper;
begin
  if AParams.Count > 1 then begin
    LHelper := gvBafDataCache.Lookup[AParams[0]];
    result := LHelper.GetCaptionGuid(AParams[1]);
    if (result = '') and (AParams.Count > 2) then
      result := AParams[2];
  end
  else
    FInter.DoLog('E', '$LOOKUP(name, value, default) - number of params less 2');
end;

function TBafVarModule.GetNumInfos(AInfo: TBafNumInfo; AParams: TStrings): string;
var
  LInt: integer;
  LCurr: currency;
begin
  result := '';
  LInt := StrToIntDef(AParams[0], MaxInt);
  LCurr := StrToCurrDef(AParams[0], MaxInt);
  case AInfo of
    niIsInt: result := IfThen(LInt = MaxInt, BAFNOCHAR, BAFYESCHAR);
    niIsCurr: result := IfThen(LCurr = MaxInt, BAFNOCHAR, BAFYESCHAR);
    niIntLen: begin
      if AParams[0] = '' then
        result := '0'
      else if LInt = MaxInt then
        result := '-1'
      else begin
        if (AParams.Count > 1) and (AnsiCompareText(AParams[1], 'sl') = 0) then
          result := IntToStr(Length(AParams[0]))
        else
          result := IntToStr(Length(IntToStr(LInt)));
      end;
    end; // case niIntLen
  end; // case
end;

function TBafVarModule.GetProc(AIndex: integer): string;
begin
  result := GetCmdList(AIndex).Text;
end;

function TBafVarModule.GetRightOfDef(AName: string): TBafRight;
var
  s: string;
begin
  result := brNone;
  if gvBafDataCache.RightIsAdmin then
    result := brWrite
  else begin
    s := FRightDefs.Values[AName] + ' ';
    case s[1] of
      'r': result := brRead;
      'w': result := brWrite;
    end;
  end;
end;

function TBafVarModule.GetCmdList(AIndex: integer): TStringList;
begin
  while AIndex > FCmdList.Count do
    FCmdList.Add(TStringList.Create);
  result := (FCmdList[AIndex - 1] as TStringList);
end;

function TBafVarModule.GetSepLine(AParam: string): string;
var
  ix: integer;
begin
  ix := StrToIntDef(AParam, -1);
  if ix >= 0 then begin
    while FSepLineValue.Count <= ix do
      FSepLineValue.Add('');
    result := FSepLineValue[ix];
  end
  else
    FInter.DoLog('E', '$SEPLINE(ix) - ix missed or below 0');
end;

function TBafVarModule.GetSepLineList(AIndex: integer): TStringList;
begin
  while AIndex > FSepLineList.Count do
    FSepLineList.Add(TStringList.Create);
  result := (FSepLineList[AIndex - 1] as TStringList);
end;

function TBafVarModule.GetTextStringList(AIndex: integer): TStringList;
begin
  while AIndex > FTextList.Count do
    FTextList.Add(TStringList.Create);
  result := (FTextList[AIndex - 1] as TStringList);
end;

function TBafVarModule.GetText(AParam: string): string;
var
  ix: integer;
begin
  ix := StrToIntDef(AParam, -1);
  if ix >= 0 then
    result := GetTextStringList(ix).Text
  else
    FInter.DoLog('E', '$TEXT(ix) - ix missing or less 0');
end;

function TBafVarModule.GetTextLine(AParams: TStrings): string;
var
  LText: TStringList;
  LParam: string;
  ix, LIndex: integer;
begin
  if AParams.Count = 0 then
    LParam := '1'
  else
    LParam := AParams[0];
  ix := StrToIntDef(LParam, 1);
  LText := GetTextStringList(ix);
  if AParams.Count > 1 then begin
    LParam := AnsiLowerCase(AParams[1]);
    if (LParam = 'cnt') or (LParam = 'count') then
      result := IntToStr(LText.Count)
    else if LParam = 'last' then begin
      if LText.Count = 0 then
        result := ''
      else
        result := LText[LText.Count - 1]
    end
    else if LParam = 'ix' then begin
      if AParams.Count > 2 then begin
        LIndex := LText.IndexOf(AParams[2]);
        if LIndex = -1 then
          result := ''
        else
          result := IntToStr(LIndex);
      end;
    end
    else begin
      ix := StrToIntDef(LParam, -1);
      if (ix = -1) or (ix >= LText.Count) then
        result := ''
      else
        result := LText[ix];
    end;
  end
  else
    result := LText.Text;
end;

function TBafVarModule.GetUserId(AParams: TStrings): string;
var
  LParam: string;
begin
  if AParams.Count > 0 then begin
    LParam := AnsiLowerCase(AParams[0]);
    if LParam = 'userguid' then
      result := dataMain.UserGuid
    else if LParam = 'userid' then
      result := dataMain.UserId;
  end;
  if result = '' then
    result := dataMain.GetDefaultUserIdent;
end;

function TBafVarModule.GetVar(AParams: TStrings): string;
var
  ix: integer;
  LInc, LVarName: string;
begin
  LVarName := AnsiLowerCase(AParams[0]);
  if Trim(LVarName) = '' then begin
    FInter.DoLog('E', '$VAR(name) - name ist leer');
    exit;
  end;
  ix := StrToIntDef(LVarName, -1234567);
  if ix = -1234567 then begin
    if AParams.Count > 1 then
    LInc := AnsiLowerCase(AParams[1]);
    if LInc = 'ib' then
      FNamedVars.Values[LVarName] := IntToStr(StrToIntDef(FNamedVars.Values[LVarName], 0) + 1);
    if LInc = 'db' then
      FNamedVars.Values[LVarName] := IntToStr(StrToIntDef(FNamedVars.Values[LVarName], 0) - 1);
    result := FNamedVars.Values[LVarName];
    if LInc = 'ia' then
      FNamedVars.Values[LVarName] := IntToStr(StrToIntDef(FNamedVars.Values[LVarName], 0) + 1);
    if LInc = 'da' then
      FNamedVars.Values[LVarName] := IntToStr(StrToIntDef(FNamedVars.Values[LVarName], 0) - 1);
  end
  else
    FInter.DoLog('E', '$VAR(name) - name is numeric');
end;

function TBafVarModule.GetVariable(AName: string): string;
begin
  result := FNamedVars.Values[AName];
end;

function TBafVarModule.ICalc(AParams: TStrings): string;
// 0 - operator
// 1 - number 1
// 2 - numer 2
// optional 3 and more: further numbers
var
  LOp: string;
  LResult: int64;
  i: integer;
begin
  if AParams.Count > 2 then begin
    LOp := AnsiLowerCase(AParams[0]);
    LResult := 0;
    if LOp = '+' then begin
      for i := 1 to AParams.Count - 1 do
        LResult := LResult + StrToInt64(AParams[i]);
    end
    else if LOp = '-' then
      LResult := StrToInt64(AParams[1]) - StrToInt64(AParams[2])
    else if LOp = '*' then
      LResult := StrToInt64(AParams[1]) * StrToInt64(AParams[2])
    else if LOp = 'div' then
      LResult := StrToInt64(AParams[1]) div StrToInt64(AParams[2])
    else if LOp = 'mod' then
      LResult := StrToInt64(AParams[1]) mod StrToInt64(AParams[2]);
    result := IntToStr(LResult);
  end
  else
    FInter.DoLog('E', '$ICALC - number of params too low');
end;

function TBafVarModule.IncDate(AParams: TStrings): string;
// 0 - date
// 1 - optional: number of days to add; default 1
var
  LDate: TDate;
  LAdd: integer;
begin
  LDate := Trunc(StrToDateTimeDef(AParams[0], 0));
  LAdd := 1;
  if AParams.Count > 1 then
    LAdd := StrToIntDef(AParams[1], 1);
  result := FormatDateTime('dd.mm.yyyy', LDate + LAdd);
end;

function TBafVarModule.IncDec(ADec: boolean; AParams: TStrings): string;
var
  LVal1, LVal2: integer;
begin
  LVal1 := StrToIntDef(AParams[0], 0);
  if AParams.Count > 1 then
    LVal2 := StrToIntDef(AParams[1], 1)
  else
    LVal2 := 1;
  if ADec then
    result := IntToStr(LVal1 - LVal2)
  else
    result := IntToStr(LVal1 + LVal2);
end;

procedure TBafVarModule.IniLoadSave(AIndex: integer; ASave: boolean);
var
  LIni: TStringIniFile;
  fn: string;
  sl: TStringList;
begin
  fn := FindParamStringReplaced('fn', '');
  LIni := GetIni(AIndex);
  sl := TStringList.Create;
  try
    if ASave then begin
      sl.Text := LIni.AsString;
      sl.SaveToFile(fn);
    end
    else begin
      sl.LoadFromFile(fn);
      LIni.AsString := sl.Text;
    end;
  finally
    sl.Free;
  end;
end;

function TBafVarModule.InOperation(AParams: TStrings; AIgnoreCase: boolean): string;
var
  LValue: string;
  i: integer;
begin
  result := BAFNOCHAR;
  if AParams.Count > 1 then begin
    LValue := AParams[0];
    for i := 1 to AParams.Count - 1 do begin
      if (AIgnoreCase and (AnsiCompareText(LValue, AParams[i]) = 0))
          or (not AIgnoreCase and (AnsiCompareStr(LValue, AParams[i]) = 0)) then begin
        result := BAFYESCHAR;
        Exit;
      end;
    end;
  end;
end;

function TBafVarModule.Int2Time(AParam: string): string;
// converts 1130 to 11:30
begin
  result := AParam;
  if Length(AParam) = 3 then
    result := '0' + AParam;
  if Length(result) = 4 then
    result := copy(result, 1, 2) + ':' + copy(result, 3, 2);
end;

function TBafVarModule.RemoveCrLf(AParams: TStrings): string;
var
  sl: TStringList;
  i: integer;
  LIns: string;
begin
  if AParams.Count > 1 then
    LIns := AParams[1];
  result := '';
  if AParams.Count > 0 then begin
    sl := TStringList.Create;
    try
      sl.Text := AParams[0];
      for i := 0 to sl.Count - 1 do
        result := result + TrimRight(sl[i]) + LIns;
    finally
      sl.Free;
    end;
  end;
end;

procedure TBafVarModule.LogVars;
var
  i: integer;
begin
  FInter.DebugLog('----------------------------------------');
  for i := 0 to FNamedVars.Count - 1 do
    FInter.DebugLog(FNamedVars[i]);
  FInter.DebugLog('----------------------------------------');
end;

procedure TBafVarModule.Loop;
var
  LRow, LLoopFrom, LLoopTo: integer;
  LBafConName, LEachRow, LName: string;
  LEachRowTrans, LNoException: boolean;
begin
  LLoopFrom := FindParamIntegerReplaced(FExecInter.LineP, 'lf', 0);
  LLoopTo := FindParamIntegerReplaced(FExecInter.LineP, 'lt', 0);
  LEachRow := FindParamString('ern', '');
  if LEachRow = '' then
    LEachRow := FindParamStringReplaced('er', '');
  LEachRowTrans := FindParamBoolean('ert', false);
  LNoException := FindParamBooleanReplaced('nex', false);
  LBafConName := FindParamStringReplacedLower('db', DC_DEFAULT);
  LName := FindParamStringReplaced('n', '');
  for LRow := LLoopFrom to LLoopTo do begin
    if LName <> '' then
      FInter.Variable[LName] := IntToStr(LRow);    // write loop var in a var
    FExecInter.EachRow(LBafConName, LEachRow, '#loop', LEachRowTrans, LNoException);
    if FInter.ProgressAborted then begin
      FInter.DoLog('I', '#loop, loop by user aborted');
      exit;
    end;
  end;
end;

function TBafVarModule.LowNoSpace(AText: string): string;
begin
  result := AnsiLowerCase(AText);
  result := StringReplace(result, ' ', '_', [rfReplaceAll]);
end;

function TBafVarModule.NewGuidFormat(AText: string): string;
begin
  result := StringReplace(AText, '{', '', []);
  result := StringReplace(result, '}', '', []);
end;

function TBafVarModule.NullValue(AParams: TStrings): string;
begin
  result := '';
  if AParams.Count > 0 then
    result := AParams[0];
  if (Trim(result) = '') and (AParams.Count > 1) then
    result := AParams[1];
end;

function TBafVarModule.OnlyNum(AParams: TStrings): string;
var
  i: integer;
  s, LMode: string;
begin
  if AParams.Count > 0 then begin
    s := AParams[0];
    if AParams.Count > 1 then
      LMode := AnsiLowerCase(AParams[1])
    else
      LMode := 'numkomma';

    for i := 1 to Length(s) do begin
      if (LMode = 'num') and CharInSet(s[i], ['0'..'9']) then
        result := result + s[i]
      else if (LMode = 'numkomma') and CharInSet(s[i], ['0'..'9', ',']) then
        result := result + s[i]
      else if (LMode = 'numpoint') and CharInSet(s[i], ['0'..'9', '.']) then
        result := result + s[i];
    end;
  end;
end;

procedure TBafVarModule.OpenText(AIndex: integer);
var
  LFileName: string;
  LText: TStringList;
  LEncoding: TEncoding;
begin
  LText := GetTextStringList(AIndex);
  LFileName := FindParamStringReplaced('fn', '');
  LEncoding := BafGetEncoding(FindParamStringReplaced('e', ''));
  if FileExists(LFileName) then
    LText.LoadFromFile(LFileName, LEncoding)
  else
    FInter.DoLog('E', '#opentext - file doesn''t exist: ' + LFileName);
end;

function TBafVarModule.FileDialog(AParams: TStrings): string;
var
  LDialog: TOpenDialog;
begin
  result := 'abort';
  if (AParams.Count > 0) and (AnsiUpperCase(AParams[0]) = 'SAVE') then
    LDialog := TSaveDialog.Create(nil)
  else
    LDialog := TOpenDialog.Create(nil);
  try
    if (AParams.Count > 1) then
      LDialog.Filter := AParams[1];
    if (AParams.Count > 2) then
      LDialog.DefaultExt := AParams[2];
    if LDialog.Execute then
      result := LDialog.FileName;
  finally
    LDialog.Free;
  end;
end;

procedure TBafVarModule.FileOperation;
var
  LOp: string;
  LResult: boolean;
begin
  if FindParamBooleanReplaced('cnd', true) then begin
    LOp := FindParamStringReplacedLower('y', '');
    LResult := false;
    if (LOp = 'cd') or (LOp = 'createdir') then
      LResult := CreateDir(FindParamStringReplaced('fn', ''))
    else if (LOp = 'rd') or (LOp = 'removedir') then
      LResult := RemoveDir(FindParamStringReplaced('fn', ''))
    else if (LOp = 'df') or (LOp = 'deletefile') then
      LResult := System.SysUtils.DeleteFile(FindParamStringReplaced('fn', ''))
    else if (LOp = 'rf') or (LOp = 'renamefile') then
      LResult := REnameFile(FindParamStringReplaced('on', ''),
          FindParamStringReplaced('fn', ''))
    else if (LOp = 'fd') or (LOp = 'forcedirectories') then
      LResult := ForceDirectories(FindParamStringReplaced('fn', ''))
    ;
    // result to value or variable
    FExecInter.SetVarOrValue('n', IfThen(LResult, 'Y', 'N'));
  end;
// procedure TBafVarModule.FileOperation
end;

procedure TBafVarModule.Forcedir;
var
  LDir: string;
begin
  LDir := FindParamStringReplaced('n', '');
  if Trim(LDir) <> '' then
    ForceDirectories(LDir)
  else
    FInter.DoLog('W', 'ForceDir not executed, dir name empty');
end;

function TBafVarModule.FormatDataTime(AParams: TStrings; ADefault: string): string;
begin
  if (AParams.Count = 0) or (AParams[0] = '') then
    result := FormatDateTime(ADefault, now)
  else
    result := FormatDateTime(AParams[0], now);
end;

function TBafVarModule.FunctionExecute(AParams: TStrings): string;
var
  LCmd, LVar: string;
begin
  if AParams.Count > 0 then begin
    LCmd := AParams[0];
    TBafInterpreterLevel.ExecInNewLevel(LCmd, FExecInter, FInter);
    if AParams.Count > 1 then
      LVar := AParams[1]
    else
      LVar := 'result';
    result := FInter.Variable[LVar];
  end;
end;

function TBafVarModule.Pad(AParams: TStrings): string;
// 0 - string
// 1 - length
// 2 - L or R
// 3 - char(s)
var
  LFolge: string;
  LLaenge, LLength: integer;
  LLinks: boolean;
begin
  if AParams.Count > 3 then begin
    result := AParams[0];
    LFolge := AParams[3];
    if LFolge = '' then
      LFolge := ' ';
    LLaenge := StrToIntDef(AParams[1], 0);
    LLinks := (UpperCase(AParams[2] + ' ')[1] = 'L');
    LLength := Length(result);
    while LLength < LLaenge do begin
      if LLinks then
        result := copy(LFolge, 1, LLaenge - LLength) + result
      else
        result := result + copy(LFolge, 1, LLaenge - LLength);
      LLength := Length(result);
    end;
  end
  else
    FInter.DoLog('E', '$PAD needs 4 parameters');
end;

procedure TBafVarModule.PasteCsv;
type
  TBafPasteMode = (pmSingle, pmMulti);
var
  LBafConName, LSep, LZeile, s, LEachRow: string;
  sl, slZeile: TStringList;
  i, p, LMax, LRowCount, LMultiRowStart: integer;
  LHasHeaderRow, LEachRowTrans, LNoException: boolean;
  LMode: TBafPasteMode;
  LExit: boolean;

  procedure lokZerlegen;
  begin
    p := Pos(LSep, LZeile);
    while p > 0 do begin
      s := copy(LZeile, 1, p - 1);
      Delete(LZeile, 1, p);
      slZeile.Add(s);
      p := Pos(LSep, LZeile);
    end;
  end; // procedure lokZerlegen

  procedure lokExecRow;
  begin
    FExecInter.EachRow(LBafConName, LEachRow, '#csv_paste', LEachRowTrans, LNoException);
    inc(LRowCount);
    if LRowCount >= LMax then begin
      FInter.DoLog('I', Format('#csv_paste, Max (%d) reached, loop aborted', [LMax]));
      LExit := true;
      exit;
    end;
    if FInter.ProgressAborted then begin
      FInter.DoLog('I', '#csv_paste, loop by user aborted');
      LExit := true;
      exit;
    end;
  end; // procedure lokExecRow

  procedure lokMulti;
  var
    i: integer;
  begin
    for i := LMultiRowStart to slZeile.Count - 1 do begin
      if (slZeile[i] <> '') and not LExit then begin
        slZeile[LMultiRowStart] := slZeile[i];
        lokExecRow;
      end;
    end;
  end; // procedure lokMulti

  procedure lokParams;
  begin
    LSep := FindParamStringReplaced('sep', #9);
    LHasHeaderRow := FindParamBooleanReplaced('hhr', false);
    s := FindParamStringLower('y', 's');
    if s = 'm' then
      LMode := pmMulti
    else
      LMode := pmSingle;
    LMax := FindParamInteger('m', MaxInt);
    LEachRow := FindParamString('ern', '');
    if LEachRow = '' then
      LEachRow := FindParamStringReplaced('er', '');
    LEachRowTrans := FindParamBoolean('ert', false);
    LNoException := FindParamBooleanReplaced('nex', false);
    LBafConName := FindParamStringReplacedLower('db', DC_DEFAULT);
    LMultiRowStart := FindParamInteger('mrs', MaxInt);
  end; // procedure lokParams

begin
  lokParams;

  slZeile := FSepLineValue;
  sl := TStringList.Create;
  try
    sl.Text := Clipboard.AsText;
    LExit := false;
    LRowCount := 0;
    for i := IfThen(LHasHeaderRow, 1, 0) to sl.Count - 1 do begin
      slZeile.Clear;
      LZeile := sl[i] + LSep;
      lokZerlegen;
      case LMode of
        pmSingle: lokExecRow;
        pmMulti: lokMulti;
      end;
      if LExit then
        exit;
    end;
  finally
    sl.Free;
  end;
// procedure TBafVarModule.PasteCsv
end;

procedure TBafVarModule.Procdir;
var
  LRec: TSearchRec;
  LRes: integer;
  LDirName, LFileName, LDone, LCommand, LName: string;
begin
  LDirName := IncludeTrailingPathDelimiter(Trim(FindParamStringReplaced('dn', '')));
  if LDirName = '' then
    Finter.DoLog('E', '#procdir - param dn empty');
  LFileName := Trim(FindParamStringReplaced('fn', '*'));
  LDone := Trim(FindParamStringReplaced('done', ''));
  LName := FindParamStringReplaced('n', 'dir_proc');
  LCommand := FindParamString('cmd', '');
  LRes := FindFirst(LDirName + LFileName, faAnyFile - faDirectory, LRec);
  try
    while LRes = 0 do begin
      FNamedVars.Values[LName] := LDirName + LRec.Name;
      TBafInterpreterLevel.ExecInNewLevel(LCommand, FExecInter, FInter);
      if DirectoryExists(LDone) then        // all executed move to done
        RenameFile(LDirName + LRec.Name, IncludeTrailingPathDelimiter(LDone) + LRec.Name);
      LRes := FindNext(LRec);
    end;
  finally
    System.SysUtils.FindClose(LRec);
  end;
end;

function TBafVarModule.ReplaceString(AParams: TStrings): string;
var
  s: string;
  LFlags: TReplaceFlags;
begin
  if AParams.Count > 2 then begin
    LFlags := [];
    if AParams.Count > 3 then begin
      s := AnsiLowerCase(AParams[3]);
      if Pos('i', s) > 0 then
        LFlags := LFlags + [rfIgnoreCase];
      if Pos('a', s) > 0 then
        LFlags := LFlags + [rfReplaceAll];
    end;
    result := StringReplace(AParams[0], AParams[1], AParams[2], LFlags);
  end
  else
    FInter.DoLog('E', '$STREP number of params to low');
end;

procedure TBafVarModule.RightsDef;
var
  i: integer;
  LName, LRight, LLine: string;
begin
  LName := FindParamStringLower('n', 'frm');
  LRight := '';
  for i := 0 to gvBafDataCache.UserGroups.Count - 1 do begin
    LLine := FindParamStringLower('r_' + gvBafDataCache.UserGroups[i], '');
    if LRight = '' then
      LRight := LLine
    else if (LRight = 'r') and (LLine = 'w') then
      LRight := 'w';
    if LRight = 'w' then
      Break;
  end;
  FRightDefs.Values[LName] := LRight;
end;

function TBafVarModule.ReplaceFunction(ACommand: string; AParams: TStrings;
  var AResult: string): boolean;
begin
  inherited;
  result := true;
  if ACommand = '$TEXT' then AResult := GetTextLine(AParams)
  else if ACommand = '$TEXTTR' then AResult := TrimRight(GetText(AParams[0]))
  else if ACommand = '$VAR' then AResult := GetVar(AParams)
  else if ACommand = '$CSV' then AResult := GetCsvValue(AParams[0], AParams[1])
  else if ACommand = '$SEPLINE' then AResult := GetSepLine(AParams[0])

  // sys and ini
  else if ACommand = '$PROG' then AResult := FInter.NeedInfo('prog', '')
  else if ACommand = '$RIGHTUSERID' then AResult := FInter.NeedInfo('rightusrid', '')
  else if ACommand = '$USERID' then AResult := FInter.NeedInfo('usrid', '')
  else if ACommand = '$INI' then AResult := GetIniValue(AParams)
  else if ACommand = '$ADM' then AResult := IsAdmin(AParams, false)
  else if ACommand = '$RADM' then AResult := IsAdmin(AParams, true)
  else if ACommand = '$INITEXT' then AResult := GetIniText(AParams[0])

  // strings
  else if ACommand = '$BASE64' then AResult := BafBase64(AParams)
  else if ACommand = '$CHR' then AResult := GetChar(AParams)
  else if ACommand = '$COPY' then AResult := StringCopy(AParams)
  else if ACommand = '$CLIPBOARD' then AResult := Clipboard.AsText
  else if ACommand = '$EXEC' then AResult := FunctionExecute(AParams)
  else if ACommand = '$GUID' then AResult := GetGuid(AParams)
  else if ACommand = '$HASH' then AResult := GetHash(AParams)
  else if ACommand = '$INVLOOKUP' then AResult := GetInversLookup(AParams)
  else if ACommand = '$LOOKUP' then AResult := GetLookup(AParams)
  else if ACommand = '$LOW' then AResult := AnsiLowerCase(AParams[0])
  else if ACommand = '$LOWNOSPACE' then AResult := LowNoSpace(AParams[0])
  else if ACommand = '$NOCRLF' then AResult := RemoveCrLf(AParams)
  else if ACommand = '$NVL' then AResult := NullValue(AParams)
  else if ACommand = '$PAD' then AResult := Pad(AParams)
  else if ACommand = '$SUBSTR' then AResult := StringCopy(AParams)
  else if ACommand = '$STREP' then AResult := ReplaceString(AParams)
  else if ACommand = '$STREXTR' then AResult := StringExtract(AParams)
  else if ACommand = '$TEXTCOMPARE' then AResult := TextCompare(AParams)
  else if ACommand = '$TRIM' then AResult := BafTrim(AParams)
  else if ACommand = '$UPP' then AResult := AnsiUpperCase(AParams[0])
  else if ACommand = '$XR' then AResult := BafXmlReplace(AParams)
  else if ACommand = '$XRR' then AResult := BafXmlReReplace(AParams)


  // integer functions, returns an integer number
  else if ACommand = '$DEC' then AResult := IncDec(true, AParams)
  else if ACommand = '$ICALC' then AResult := ICalc(AParams)
  else if ACommand = '$INC' then AResult := IncDec(false, AParams)
  else if ACommand = '$INTLEN' then AResult := GetNumInfos(niIntLen, AParams)
  else if ACommand = '$LEN' then AResult := IntToStr(Length(AParams[0]))
  else if ACommand = '$POS' then AResult := BafPos(AParams)
  else if ACommand = '$RANDOM' then AResult := BafRandom(AParams)

  // currency functions, retuns a currency
  else if ACommand = '$CCALC' then AResult := CCalc(AParams)
  else if ACommand = '$CURR' then AResult := GetCurr(AParams)
  else if ACommand = '$ONLYNUM' then AResult := OnlyNum(AParams)

  // date functions, retuns a date or a time
  else if ACommand = '$CONVERT' then AResult := Convert(AParams)
  else if ACommand = '$INCDATE' then AResult := IncDate(AParams)
  else if ACommand = '$INT2TIME' then AResult := Int2Time(AParams[0])
  else if ACommand = '$NOW' then AResult := FormatDateTime('dd.mm.yyyy hh:mm:ss', now)
  else if ACommand = '$NOWFMT' then AResult := FormatDateTime(AParams[0], now)
  else if ACommand = '$NOWMIN' then AResult := FormatDateTime('dd.mm.yyyy hh:mm', now)
  else if ACommand = '$TODAY' then AResult := FormatDataTime(AParams, 'dd.mm.yyyy')
  else if ACommand = '$DFMT' then AResult := GetDateFormated(AParams)

  // boolean functions, retuns Y or N
  else if ACommand = '$BCALC' then AResult := BCalc(AParams)
  else if ACommand = '$BOOL' then AResult := IfThen(FExecInter.ParseBoolStatement(AParams[0]), BAFYESCHAR, BAFNOCHAR)
  else if ACommand = '$EMPTY' then AResult := IfThen(Trim(AParams[0]) = '', BAFYESCHAR, BAFNOCHAR)
  else if ACommand = '$ISCURR' then AResult := GetNumInfos(niIsCurr, AParams)
  else if ACommand = '$ISINT' then AResult := GetNumInfos(niIsInt, AParams)
  else if ACommand = '$IN' then AResult := InOperation(AParams, false)
  else if ACommand = '$INIC' then AResult := InOperation(AParams, true)
  else if ACommand = '$INTMAX' then AResult := GetIntMinMax(niIntMax, AParams[0], AParams[1])
  else if ACommand = '$INTMIN' then AResult := GetIntMinMax(niIntMin, AParams[0], AParams[1])
  else if ACommand = '$NEMPTY' then AResult := IfThen(Trim(AParams[0]) = '', BAFNOCHAR, BAFYESCHAR)
  else if ACommand = '$NUMBETWEEN' then AResult := GetBetween(niNumBetween, AParams)
  else if ACommand = '$DATEBETWEEN' then AResult := GetBetween(niDateBetween, AParams)
  else if ACommand = '$DATEMINREL' then AResult := GetDateMinMaxRel(niDateMin, AParams[0], AParams[1])
  else if ACommand = '$DATEMAXREL' then AResult := GetDateMinMaxRel(niDateMax, AParams[0], AParams[1])
  else if ACommand = '$NOT' then AResult := IfThen(BafIsYesChar(AParams[0]), BAFNOCHAR, BAFYESCHAR)
  else if ACommand = '$REGEX' then AResult := BafRegex(AParams)
  else if ACommand = '$TEXT_HASLINE' then AResult := TextHasLine(AParams)
  else if ACommand = '$VIBAN' then AResult := IfThen(BafIsValidIban(AParams[0]), BAFYESCHAR, BAFNOCHAR)

  // files and directories
  else if ACommand = '$DIR' then AResult := BafGetDir(AParams[0])
  else if ACommand = '$FILEDIALOG' then AResult := FileDialog(AParams)



  // colors
  else if ACommand = '$CLRDEADLINE_YR' then AResult := GetClrDeadline(ciDeadlineYR, AParams)


  // JSON and XML
  else if ACommand = '$JSON' then AResult := GetJSON(AParams)

  else result := false;


end;

procedure TBafVarModule.Save2Clipboard;
begin
  Clipboard.AsText := FindParamStringReplaced('z', '');
end;

procedure TBafVarModule.SaveText(AIndex: integer);
var
  fn: string;
  LSort, LWriteBom, LBomAlt: boolean;
  LText, sl: TStringList;
  i: integer;
  LEncoding: TEncoding;
begin
  LText := GetTextStringList(AIndex);
  fn := FindParamStringReplaced('fn', '');
  LEncoding := BafGetEncoding(FindParamStringReplaced('e', ''));
  LWriteBom := FindParamBooleanReplaced('bom', true);
  LSort := FindParamBoolean('sort', false);
  if LSort then begin
    sl := TStringList.Create;
    try
      sl.Sorted := true;
      sl.Duplicates := dupIgnore;
      for i := 0 to LText.Count - 1 do
        sl.Add(LText[i]);
      sl.WriteBOM := LWriteBom;
      sl.SaveToFile(fn, LEncoding);
    finally
      sl.Free;
    end;
  end
  else begin
    LBomAlt := LText.WriteBOM;
    LText.WriteBOM := LWriteBom;
    LText.SaveToFile(fn, LEncoding);
    LText.WriteBOM := LBomAlt;
  end;
  if FindParamBooleanReplaced('o', false) then
    BafOpenFile(fn);
end;

procedure TBafVarModule.SepLine(AIndex: integer);
var
  LRow, LRowCount, LMax: integer;
  LBafConName, LSep, LEachRow: string;
  LEachRowTrans, LNoException: boolean;
  LText: TStringList;
begin
  LText := GetSepLineList(AIndex);
  LSep := FindParamString('sep', ';') + ' ';
  LText.Delimiter := LSep[1];
  LText.DelimitedText := FindParamStringReplaced('z', '');
  LMax := FindParamIntegerReplaced('m', MaxInt);
  LEachRow := FindParamString('ern', '');
  if LEachRow = '' then
    LEachRow := FindParamStringReplaced('er', '');
  LEachRowTrans := FindParamBooleanReplaced('ert', false);
  LNoException := FindParamBooleanReplaced('nex', false);
  LBafConName := FindParamStringReplacedLower('db', DC_DEFAULT);
  if (LEachRow <> '') and (LText.Count > 0) then begin
    LRowCount := 0;
    for LRow := 0 to LText.Count - 1 do begin
      while FSepLineValue.Count <= AIndex do
        FSepLineValue.Add('');
      FSepLineValue[AIndex] := Trim(LText[LRow]);
      FExecInter.EachRow(LBafConName, LEachRow, '#sepline', LEachRowTrans, LNoException);
      inc(LRowCount);
      if LRowCount >= LMax then begin
        FInter.DoLog('I', Format('#sepline, Max (%d) reached, loop aborted', [LMax]));
        Break;
      end;
      if FInter.ProgressAborted then begin
        FInter.DoLog('I', '#sepline, loop by user aborted');
        Break;
      end;
    end;
  end;
// procedure SepLine.SepLine
end;

procedure TBafVarModule.SetCmd(AIndex: integer);
begin
  GetCmdList(AIndex).Add(FExecInter.LineP);
end;

procedure TBafVarModule.SetIniValue(AIndex: integer);
var
  LIni: TStringIniFile;
  s: string;
begin
  s := FindParamStringLower('n', '');
  if (s = 'usr') or (s = 'user') then
    LIni := dataMain.UserIni
  else
    LIni := GetIni(AIndex);
  LIni.WriteString(FindParamStringReplaced('sec', 'values'),
      FindParamStringReplaced('i', 'value'),
      FindParamStringReplaced('z', ''))
end;

procedure TBafVarModule.SetText(AIndex: integer);
var
  s: string;
begin
  s := FExecInter.ReplaceFunctions(FExecInter.LineP);
  s := FExecInter.ReplaceFunctions(s);
  GetTextStringList(AIndex).Add(s);
end;

procedure TBafVarModule.SetTextN(AIndex: integer);
begin
  GetTextStringList(AIndex).Add(FExecInter.LineP);
end;

procedure TBafVarModule.SetVar(AOnlyIfEmpty: boolean);
var
  LName, LValue, LRightName: string;
  LRight: TBafRight;
begin
  if FindParamBooleanReplaced('cnd', true) then begin
    LRightName := FindParamStringLower('r', 'frm');
    LName := FindParamStringReplaced('n', '');
    if LRightName = 'frm' then
      LRight := brWrite
    else
      LRight := FInter.GetRightOfDef(LRightName);
    case LRight of
      brNone: LValue := FindParamStringReplaced('zn', '');
      brRead: LValue := FindParamStringReplaced('zr', '');
    else
      LValue := FindParamStringReplaced('z', '');
    end;
    if FindParamBooleanReplaced('rf', false) then
      LValue := FExecInter.ReplaceFunctions(LValue);
    if LValue = '' then
      LValue := FindParamStringReplaced('ie', '');
    if (AOnlyIfEmpty = false) or (Trim(FNamedVars.Values[LName]) = '') then
      FNamedVars.Values[LName] := LValue;
  end;
end;

procedure TBafVarModule.SetVariable(AName: string; const Value: string);
begin
  inherited;
  FNamedVars.Values[AName] := Value;
end;

function TBafVarModule.StringCopy(AParams: TStrings): string;
var
  s: string;
  LVon, LCount: integer;
begin
  result := '';
  if AParams.Count > 2 then begin
    s := AParams[0];
    LVon := StrToIntDef(AParams[1], 1);
    LCount := StrToIntDef(AParams[2], MaxInt);
    result := copy(s, LVon, LCount);
  end
  else if AParams.Count = 2 then begin
    s := AParams[0];
    LVon := StrToIntDef(AParams[1], 1);
    result := copy(s, LVon, MaxInt);
  end;
end;

function TBafVarModule.StringExtract(AParams: TStrings): string;
var
  LText, LType: string;

  function lokFirst(AParamNum: integer; AFrom, AInclude: boolean): string;
  var
    LPos: integer;
    LSearch: string;
  begin
    if AParams.Count > AParamNum then begin
      LSearch := AParams[AParamNum];
      LPos := Pos(LSearch, LText);
      if AFrom then begin
        if LPos > 0 then
          result := copy(LText, LPos
              + IfThen(AInclude, 0, Length(LSearch)), MaxInt)
        else
          result := '';
      end
      else begin
        if LPos > 0 then
          result := copy(LText, 1, LPos - 1
              + IfThen(not AInclude, 0, Length(LSearch)))
        else
          result := LText;
      end;
    end
    else
      FInter.DoLog('E', '$STREXTR - number of params too low for ' + LType);
  end; // function lokFirst

begin
  if AParams.Count > 2 then begin
    LText := AParams[0];
    LType := Trim(AnsiLowerCase(AParams[1]));
    if LType = 'ffi' then
      result := lokFirst(2, true, true)
    else if LType = 'ffe' then
      result := lokFirst(2, true, false)
    else if LType = 'tfi' then
      result := lokFirst(2, false, true)
    else if LType = 'tfe' then
      result := lokFirst(2, false, false)


    ;
  end
  else
    FInter.DoLog('E', '$STREXTR - number of params too low');
// function TBafVarModule.StringExtract
end;

function TBafVarModule.TextCompare(AParams: TStrings): string;
var
  LText0, LText1: string;
begin
  if AParams.Count > 1 then begin
    LText0 := Trim(GetTextStringList(StrToIntDef(AParams[0], 1)).Text);
    LText1 := Trim(GetTextStringList(StrToIntDef(AParams[1], 1)).Text);
    result := IfThen(LText0 = LText1, 'Y', 'N');
  end;
end;

procedure TBafVarModule.TextCompose(AIndex: integer);
var
  LText: TStringList;
  i, ix: integer;
  LName, LSep, LResult: string;
  LLastSep: boolean;
begin
  if FindParamBooleanReplaced('cnd', true) then begin
    LText := GetTextStringList(AIndex);
    LSep := FindParamStringReplaced('sep', '');
    LLastSep := FindParamBooleanReplaced('ls', false);
    for i := 0 to LText.Count - 1 do
      LResult := LResult + LText[i]
          + IfThen(LLastSep or (i < (LText.Count - 1)), LSep, '');
    FExecInter.SetVarOrValue('n', LResult);
  end;
end;

procedure TBafVarModule.TextDissect(AIndex: integer);
var
  LText: TStringList;
  p: integer;
  LValue, LSep: string;
begin
  if FindParamBooleanReplaced('cnd', true) then begin
    LText := GetTextStringList(AIndex);
    LValue := FindParamStringReplaced('z', '');
    LSep := FindParamStringReplaced('sep', '');
    if FindParamBooleanReplaced('ls', false) then
      LValue := LValue + LSep;
    p := Pos(LSep, LValue);
    while p > 0 do begin
      LText.Add(copy(LValue, 1, p - 1));
      Delete(LValue, 1, p - 1 + Length(LSep));
      p := Pos(LSep, LValue);
    end;
  end;
end;

function TBafVarModule.TextHasLine(AParams: TStrings): string;
// 0 - Num of Text
// 1 - Value to check
// 2 - Yes
// 3 - No
var
  LText: TStringList;
  LYes, LNo: string;
begin
  if AParams.Count > 1 then begin
    LText := GetTextStringList(StrToIntDef(AParams[0], 1));
    LYes := 'Y';
    LNo := 'N';
    if AParams.Count > 2 then
      LYes := AParams[2];
    if AParams.Count > 3 then
      LNo := AParams[3];
    result := IfThen(LText.IndexOf(AParams[1]) >= 0, LYes, LNo);
  end
  else
    FInter.DoLog('E', '$TEXT_HASLINE - number of params too low');
end;

procedure TBafVarModule.TextLoop(AIndex: integer);
var
  LText: TStringList;
  i, LNum, LMax, LRowCount: integer;
  LName, LEachRow, LBafConName: string;
  LEachRowTrans, LNoException: boolean;
begin
  if FindParamBooleanReplaced('cnd', true) then begin
    LMax := FindParamIntegerReplaced('m', MaxInt);
    LEachRow := FindParamString('ern', '');
    if LEachRow = '' then
      LEachRow := FindParamStringReplaced('er', '');
    LEachRowTrans := FindParamBooleanReplaced('ert', false);
    LNoException := FindParamBooleanReplaced('nex', false);
    LBafConName := FindParamStringReplacedLower('db', DC_DEFAULT);
    LName := FindParamStringReplaced('n', '');
    LNum := StrToIntDef(LName, -1);
    LText := GetTextStringList(AIndex);
    LRowCount := 0;
    for i := 0 to LText.Count - 1 do begin
      if LNum > 0 then
        FExecInter.Values[LNum] := IntToStr(i)
      else if LName <> '' then
        FInter.Variable[LName] := IntToStr(i);

      FExecInter.EachRow(LBafConName, LEachRow, FExecInter.LineF, LEachRowTrans, LNoException);
      inc(LRowCount);
      if LRowCount >= LMax then begin
        FInter.DoLog('I', Format(FExecInter.LineF + ', Max (%d) reached, execution aborted', [LMax]));
        Break;
      end;
      if FInter.ProgressAborted then begin
        FInter.DoLog('I', FExecInter.LineF + ', loop by user aborted');
        Break;
      end;
    end;
  end;
// procedure TBafVarModule.TextLoop
end;

procedure TBafVarModule.TextProperties(AIndex: integer);
var
  LText: TStringList;
  s: string;
begin
  LText := GetTextStringList(AIndex);
  LText.Sorted := FindParamBooleanReplaced('srt', false);
  s := FindParamStringReplacedLower('dup', '');
  if s = 'ign' then
    LText.Duplicates := dupIgnore
  else if s = 'acc' then
    LText.Duplicates := dupAccept
  else if s = 'err' then
    LText.Duplicates := dupError;
end;

procedure TBafVarModule.TextSet(AIndex: integer);
var
  LText: TStringList;
  s, LValue: string;
  ix: integer;
begin
  LText := GetTextStringList(AIndex);
  s := FindParamStringReplacedLower('i', '');
  LValue := FindParamStringReplaced('z', '');
  if s = 'all' then
    LText.Text := LValue
  else begin
    if s = 'last' then
      ix := LText.Count - 1
    else
      ix := StrToIntDef(s, -1);
    if (ix = -1) or (ix >= LText.Count) then
      FInter.DoLog('W', '#text_set - ix out of range')
    else
      LText[ix] := LValue;
  end;
end;

procedure TBafVarModule.TextSort(AIndex: integer);
var
  LText, sl: TStringList;
  LType: string;
  i: integer;
begin
  LText := GetTextStringList(AIndex);
  LType := FindParamStringReplacedLower('y', '');
  if LType = 'inv' then begin
    sl := TStringList.Create;
    try
      for i := LText.Count - 1 downto 0 do
        sl.Add(LText[i]);
      LText.Text := sl.Text;
    finally
      sl.Free;
    end;
  end
  else
    LText.Sort;
end;

function TBafVarModule.InterpretLine(AExecInter: TBafCustomInterpreter): boolean;
var
  LNum: integer;
  LInter: TBafCustomInterpreter;
begin
  LInter := FExecInter;
  try
    FExecInter := AExecInter;

    result := true;
    if FExecInter.LineF = '#text_clearall' then FTextList.Clear                              // clear all texts
    else if FExecInter.LineF = '#textclearall' then FTextList.Clear
    else if BafIsNumberedFunk(FExecInter.LineF, '#text', LNum) then SetText(LNum)            // add a line to the text
    else if BafIsNumberedFunk(FExecInter.LineF, '#textn', LNum) then SetTextN(LNum)
    else if BafIsNumberedFunk(FExecInter.LineF, '#text_clear', LNum) then ClearText(LNum)    // clears one text
    else if BafIsNumberedFunk(FExecInter.LineF, '#textclear', LNum) then ClearText(LNum)
    else if BafIsNumberedFunk(FExecInter.LineF, '#text_save', LNum) then SaveText(LNum)      // saves the text to a file
    else if BafIsNumberedFunk(FExecInter.LineF, '#textsave', LNum) then SaveText(LNum)
    else if BafIsNumberedFunk(FExecInter.LineF, '#text_open', LNum) then OpenText(LNum)      // opens the text from a file
    else if BafIsNumberedFunk(FExecInter.LineF, '#textopen', LNum) then OpenText(LNum)
    else if BafIsNumberedFunk(FExecInter.LineF, '#text_props', LNum) then TextProperties(LNum)
    else if BafIsNumberedFunk(FExecInter.LineF, '#text_set', LNum) then TextSet(LNum)
    else if BafIsNumberedFunk(FExecInter.LineF, '#text_compose', LNum) then TextCompose(LNum)
    else if BafIsNumberedFunk(FExecInter.LineF, '#text_dissect', LNum) then TextDissect(LNum)
    else if BafIsNumberedFunk(FExecInter.LineF, '#text_sort', LNum) then TextSort(LNum)
    else if BafIsNumberedFunk(FExecInter.LineF, '#text_loop', LNum) then TextLoop(LNum)

    else if BafIsNumberedFunk(FExecInter.LineF, '#ini_val', LNum) then SetIniValue(LNum)
    else if BafIsNumberedFunk(FExecInter.LineF, '#ini_load', LNum) then IniLoadSave(LNum, false)
    else if BafIsNumberedFunk(FExecInter.LineF, '#ini_save', LNum) then IniLoadSave(LNum, true)


    else if BafIsNumberedFunk(FExecInter.LineF, '#cmd', LNum) then SetCmd(LNum)
    else if BafIsNumberedFunk(FExecInter.LineF, '#cmd_clear', LNum) then ClearCmd(LNum)
    else if BafIsNumberedFunk(FExecInter.LineF, '#cmdclear', LNum) then ClearCmd(LNum)
    else if FExecInter.LineF = '#cmd_clearall' then ClearCmds
    else if FExecInter.LineF = '#cmdclearall' then ClearCmds
    else if FExecInter.LineF = '#code' then FCode.Add(FExecInter.LineP)                 // adds a code line
    else if FExecInter.LineF = '#setvar' then SetVar(false)
    else if FExecInter.LineF = '#var_set' then SetVar(false)                            // sets a variable
    else if FExecInter.LineF = '#var_clearall' then FNamedVars.Clear                    // clears all variables
    else if FExecInter.LineF = '#setvarempty' then SetVar(true)
    else if FExecInter.LineF = '#var_setempty' then SetVar(true)                        // sets a variable if empty
    else if FExecInter.LineF = '#var_add' then AddVar                                   // adds to a variable
    else if FExecInter.LineF = '#var_log' then LogVars                                  // writes all variables to the log
    else if BafIsNumberedFunk(FExecInter.LineF, '#sepline', LNum) then SepLine(LNum)    // converts a string in a sparated list

    else if FExecInter.LineF = '#rights' then RightsDef
    else if FExecInter.LineF = '#rights_clear' then FRightDefs.Clear

    else if BafIsNumberedFunk(FExecInter.LineF, '#csv_open', LNum) then CsvOpen(LNum)   // opens a csv file
    else if BafIsNumberedFunk(FExecInter.LineF, '#csv_line', LNum) then CsvLine(LNum)   // executes a command for each line of the csv file
    else if FExecInter.LineF = '#csv_paste' then PasteCsv                               // execute csv data in the clipboard

    else if FExecInter.LineF = '#exec' then Exec                                        // executes a command
    else if FExecInter.LineF = '#dir_proc' then Procdir                                 // processes all files in a directory
    else if FExecInter.LineF = '#dir_force' then Forcedir                               // forces a dir
    else if FExecInter.LineF = '#batchexec' then BatchExec                              // converts a text into a batch and executes it
    else if FExecInter.LineF = '#openfile' then DoOpenFile
    else if FExecInter.LineF = '#file_open' then DoOpenFile                             // WIN-Api OpenFile
    else if FExecInter.LineF = '#clipboard' then Save2Clipboard                         // copy data to the clipboard
    else if FExecInter.LineF = '#file_op' then FileOperation                            // files and dirs
    else if FExecInter.LineF = '#file_search' then BafFileSearch                        // searches files

    else if FExecInter.LineF = '#loop' then Loop                                        // processes a loop

    else if FExecInter.LineF = '#check_string' then CheckString

    else result := false;
  finally
    FExecInter := LInter;
  end;
end;

function TBafVarModule.IsAdmin(AParams: TStrings; ARight: boolean): string;
var
  AResult: boolean;
begin
  if ARight then
    AResult := gvBafDataCache.RightIsAdmin
  else
    AResult := gvBafDataCache.IsAdmin;
  if AResult then begin
    if AParams.Count > 0 then
      result := AParams[0]
    else
      result := BAFYESCHAR
  end
  else begin
    if AParams.Count > 1 then
      result := AParams[1]
    else
      result := BAFNOCHAR
  end;
end;

end.

