unit uBafXlsModule;

// this code is under the BAF fair use license (BFUL) - https://bafbal.de/index.php?title=Bful
// exporting in xlsx- format
// needs http://www.kluug.net/xlsx-ods-delphi.php
// https://bafbal.de/index.php?title=Modul_XLS


interface

uses System.Math, System.SysUtils, System.Classes, uBafTypes, uBafInterpreter,
  OExport, OExport_FMX, System.UITypes, uBafPage;

type
  TBafXlsModule = class(TBafInterpreterCustomModule)
  protected
    FExport: TOExport;
    FSheet: TExportWorkSheet;
    FSheetRow: TExportRow;
    FExportInvisibleColumns: boolean;
    procedure XlsStart;
    procedure XlsStop;
    procedure XlsAddSheet;
    procedure XlsAddRow;
    procedure XlsAddCell;
    procedure XlsPage;
    function ConvertAlign(AAlign: TBafAlignment): TCellHAlignment;
    function AddCellTyp(AType: TBafPageCellType; AValue: string): TExportCell;
    procedure ExportSimpleGrid(AGrid: TBafSimpleGrid);
    procedure ExportRows(AGrid: TBafSimpleGrid; AType: TBafGridRowType);
    function ExportSegment(ASegment: TBafPageSegment): boolean;
  public
    constructor Create; override;
    destructor Destroy; override;
    function InterpretLine(AExecInter: TBafCustomInterpreter): boolean; override;
    function ReplaceFunction(ACommand: string; AParams: TStrings;
        var AResult: string): boolean; override;
    procedure ExportSegmentXls(ASegment: TObject); override;
    procedure ExportPageXls(APage: TObject); override;
  end;


implementation

uses foBafDialog, foMain;

{ TBafXlsModule }

function TBafXlsModule.AddCellTyp(AType: TBafPageCellType; AValue: string): TExportCell;
var
  LDate: TDateTime;
begin
  case AType of
    ctDate, ctDateMin, ctDateSek: begin
      LDate := StrToDateTimeDef(AValue, -1000001);
      if LDate > -1000000 then begin
        case AType of
          ctDate: result := FSheetRow.AddCellDate(LDate);
          ctDateMin: result := FSheetRow.AddCellDateTimeWS(LDate);
          ctDateSek: result := FSheetRow.AddCellDateTime(LDate);
        end;
      end
      else
        result := FSheetRow.AddCellEmpty;
    end;
    ctInt: result := FSheetRow.AddCellNumber(StrToIntDef(AValue, 0), 0);
    ctCurr: result := FSheetRow.AddCellNumber(StrToCurrDef(AValue, 0), 2);
    ctCurr4: result := FSheetRow.AddCellNumber(StrToCurrDef(AValue, 0), 4);
    else
      result := FSheetRow.AddCellString(AValue);
  end;
end;

function TBafXlsModule.ConvertAlign(AAlign: TBafAlignment): TCellHAlignment;
begin
  case AAlign of
    taRightJustify: result := cahRight;
    taCenter: result := cahCenter;
    taExplicitLeft: result := cahLeft;
    taDecimal2, taDecimal4: result := cahJustify;
  else
    result := cahAuto;
  end;
end;

constructor TBafXlsModule.Create;
begin
  inherited;

end;

destructor TBafXlsModule.Destroy;
begin

  inherited;
end;

procedure TBafXlsModule.ExportPageXls(APage: TObject);
var
  LPage: TBafPage;
begin
  inherited;
  if Assigned(FExport) then
    FreeAndNil(FExport);
  FExport := TOExport.Create;
  try
    LPage := (APage as TBafPage);
    LPage.AllSegFunction(ExportSegment);
    FExport.SaveToFileWithDialog('baf_export', '', true);
  finally
    FreeAndNil(FExport);
  end;
end;

procedure TBafXlsModule.ExportRows(AGrid: TBafSimpleGrid; AType: TBafGridRowType);
var
  LRow, LCol, LColSpan: integer;
  LCell: TBafSgCell;
  LColumn: TBafSgColumn;
  LXlsCell: TExportCell;
  LWidth: single;
begin
  for LRow := 0 to AGrid.RowCount(AType) - 1 do begin
    FSheetRow := FSheet.AddRow;
    LColSpan := 0;
    for LCol := 0 to AGrid.Columns.Count - 1 do begin
      LCell := AGrid.Cells[AType, LCol, LRow];
      LColumn := LCell.Parents.Column;
      LWidth := FindParamSingleReplaced(LColumn.LineP, 'xw', -1);
      if LWidth < 0 then
        LWidth := LColumn.Width;
      if LWidth > 0 then begin
        if LColSpan <= 0 then begin
          if AType = rtHeader then
            LXlsCell := FSheetRow.AddCellString(LCell.Text)
          else if (AType = rtFooter) and (LCell.GetDisplayText = '') then
            LXlsCell := FSheetRow.AddCellString('')
          else if LColumn.CellType in [ctGuid] then
            LXlsCell := AddCellTyp(LColumn.CellType, LCell.Text)
          else
            LXlsCell := AddCellTyp(LColumn.CellType, LCell.GetDisplayText);
          LXlsCell.Width := LWidth;
          if (AType = rtHeader) then begin
            LColSpan := FindParamIntegerReplaced(LColumn.LineP,
                'xcs' + IntToStr(LRow + 1), 1);
            LXlsCell.ColSpan := LColSpan;
            LXlsCell.BGColor := FInter.FindParamColor(LColumn.LineP, 'xhcl', $FFCCBB);
            LXlsCell.SetBorders(cbAll, ebThin);
          end;
        end;
      end;
      dec(LColSpan);
    end;
  end;
end;

function TBafXlsModule.ExportSegment(ASegment: TBafPageSegment): boolean;
begin
  result := true;
  if ASegment.XlsCommand = '' then begin
    case ASegment.SegmentType of
      stValueList, stGrid, stXGrid, stXXGrid: ExportSimpleGrid(ASegment.Grid);

    end;
  end
  else
    TBafInterpreterLevel.ExecInNewLevel(ASegment.XlsCommand, FExecInter, FInter);
end;

procedure TBafXlsModule.ExportSegmentXls(ASegment: TObject);
begin
  inherited;
  if Assigned(FExport) then
    FreeAndNil(FExport);
  FExport := TOExport.Create;
  try
    ExportSegment(ASegment as TBafPageSegment);
    FExport.SaveToFileWithDialog('baf_export', '', true);
  finally
    FreeAndNil(FExport);
  end;
end;

procedure TBafXlsModule.ExportSimpleGrid(AGrid: TBafSimpleGrid);
begin
  FSheet := FExport.AddWorkSheet(AGrid.Parents.Segment.SegmentName);
  ExportRows(AGrid, rtHeader);
  ExportRows(AGrid, rtDisplayData);
  ExportRows(AGrid, rtFooter);
end;

function TBafXlsModule.InterpretLine(AExecInter: TBafCustomInterpreter): boolean;
var
  LNum: integer;
  LInter: TBafCustomInterpreter;
begin
  result := false;
  LInter := FExecInter;
  try
    FExecInter := AExecInter;

    result := true;
    if FExecInter.LineF = '#xls_start' then XlsStart                            // start a calc sheet
    else if FExecInter.LineF = '#xls_stop' then XlsStop                         // ends a calc sheet
    else if FExecInter.LineF = '#xls_sheet' then XlsAddSheet                    // adds a sheet
    else if FExecInter.LineF = '#xls_row' then XlsAddRow                        // adds a row
    else if FExecInter.LineF = '#xls_cell' then XlsAddCell                      // adds a cell

    else if FExecInter.LineF = '#xls_page' then XlsPage                         // exports a BAF Page

    else result := false;
  finally
    FExecInter := LInter;
  end;


end;

function TBafXlsModule.ReplaceFunction(ACommand: string; AParams: TStrings;
  var AResult: string): boolean;
begin
  inherited;
  result := true;
  if ACommand = '$XLS_NONE' then
    AResult := ''
  else result := false;
end;

procedure TBafXlsModule.XlsAddCell;
var
  LCell: TExportCell;
  LValue: string;
  LType: TBafPageCellType;
  LWidth: single;
  LAlign: TBafAlignment;
  LColor: TColor;
begin
  if FindParamBooleanReplaced('cnd', true) then begin
    LValue := FindParamStringReplaced('z', '');
    LType := BafGetCellType(FindParamStringLower('y', ''));
    LCell := AddCellTyp(LType, LValue);
    LWidth := FindParamSingleReplaced('w', -1);
    if LWidth >= 0 then
      LCell.SetWidth(LWidth);
    LAlign := FInter.FindParamAlignment(FExecInter.LineP, 'a', taLeftJustify);
    LCell.SetAlignment(ConvertAlign(LAlign));
    LColor := FindParamColor('cl', clWhite);
    if LColor <> clWhite then
      LCell.SetBGColor(LColor);
    LCell.SetFontColor(FindParamColor('fc', clBlack));
  end;
end;

procedure TBafXlsModule.XlsAddRow;
begin
  if FindParamBooleanReplaced('cnd', true) then begin
    FSheetRow := FSheet.AddRow;
  end;
end;

procedure TBafXlsModule.XlsAddSheet;
begin
  if FindParamBooleanReplaced('cnd', true) then begin
    FSheet := FExport.AddWorkSheet(FindParamStringReplaced('c', ''));
    FSheet.WindowSettings.Split.Row := FindParamIntegerReplaced('fr', 0);
    FSheet.WindowSettings.Split.Col := FindParamIntegerReplaced('fc', 0);
  end;
end;

procedure TBafXlsModule.XlsPage;
var
  LPage: TBafPage;
  LPrim, LCat, LSeg: integer;
  LCategory: TBafPageCategory;
  LSegment: TBafPageSegment;
begin
  if FindParamBooleanReplaced('cnd', true) then begin
    if Assigned(FExport) then
      FreeAndNil(FExport);
    FExport := TOExport.Create;
    try
      LPage := FExecInter.GetFormObject(foPage) as TBafPage;
      for LPrim := 0 to LPage.PrimaryDivs.Count - 1 do begin
        for LCat := 0 to LPage.PrimaryDivs.Items[LPrim].Categories.Count - 1 do begin
          LCategory := LPage.PrimaryDivs.Items[LPrim].Categories.Items[LCat];
          for LSeg := 0 to LCategory.Segments.Count - 1 do begin
            LSegment := LCategory.Segments.Items[LSeg];
            ExportSegment(LSegment);
          end;
        end;
      end;
      FExport.SaveToFileWithDialog('baf_export', '', true);
    finally
      FreeAndNil(FExport);
    end;
  end;
end;

procedure TBafXlsModule.XlsStart;
begin
  if FindParamBooleanReplaced('cnd', true) then begin
    if Assigned(FExport) then
      FreeAndNil(FExport);
    FExport := TOExport.Create;
    FExportInvisibleColumns := FindParamBooleanReplaced('eic', false);
  end;
end;


procedure TBafXlsModule.XlsStop;
var
  LOpen: boolean;
  LFilename: string;
begin
  if FindParamBooleanReplaced('cnd', true) then begin
    LOpen := FindParamBooleanReplaced('o', true);
    LFilename := FindParamStringReplaced('fn', '');
    if LFilename = '' then
      FExport.SaveToFileWithDialog('baf_export', '', LOpen)
    else
      FExport.SaveToFile(LFilename, LOpen);
    FreeAndNil(FExport);
  end;
end;

initialization
  OExportDateFormat := 'dd.mm.yyyy';
  OExportTimeFormat := 'hh:mm:ss';

end.
